# Copyright 2015 Brocade Communications System, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import codecs
import os
import yaml

from tacker import context
from tacker.extensions import vnfm
from tacker import objects
from tacker.tests.unit import base
from tacker.tests import uuidsentinel
from tacker.vnfm.infra_drivers.openstack.translate_template import TOSCAToHOT


class TestEtsiTranslateTemplate(base.TestCase):

    def setUp(self):
        super(TestEtsiTranslateTemplate, self).setUp()
        self.tth = TOSCAToHOT(None, None)
        self.tth.fields = {}
        self.tth.vnf = {}
        self.tth.vnf['attributes'] = {}

    def _get_template(self, name):
        filename = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), name)
        with codecs.open(filename, encoding='utf-8', errors='strict') as f:
            return f.read()

    def _load_yaml(self, yaml_name, update_import=False):
        filename = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), yaml_name)
        file_abspath = os.path.dirname(os.path.abspath(filename))
        with open(filename, 'r') as f:
            heat_yaml = f.read()
            heat_dict = yaml.safe_load(heat_yaml)
            if update_import:
                self._update_imports(heat_dict, file_abspath)
            return heat_dict

    def _update_imports(self, yaml_dict, file_abspath):
        imports = yaml_dict['imports']
        new_imports = []
        for i in imports:
            new_imports.append(file_abspath + '/' + i)
        yaml_dict['imports'] = new_imports

    def test_generate_hot_from_tosca(self):
        tosca_file = './data/etsi_nfv/' \
            'tosca_generate_hot_from_tosca.yaml'
        hot_file = './data/etsi_nfv/hot/' \
            'hot_generate_hot_from_tosca.yaml'
        vnfd_dict = self._load_yaml(tosca_file, update_import=True)

        # Input params
        dev_attrs = {}

        data = [{
            "id": 'VL1',
            "resource_id": 'neutron-network-uuid_VL1',
                "ext_cps": [{
                    "cpd_id": "CP1",
                    "cp_config": [{
                        "cp_protocol_data": [{
                            "layer_protocol": "IP_OVER_ETHERNET",
                            "ip_over_ethernet": {
                                "mac_address": 'fa:16:3e:11:11:11',
                                "ip_addresses": [{
                                    'type': 'IPV4',
                                    'fixed_addresses': ['1.1.1.1'],
                                    'subnet_id': 'neutron-subnet-uuid_CP1'}]}
                        }]
                    }]}]},
               {
                "id": 'VL2',
                "resource_id": 'neutron-network-uuid_VL2',
                "ext_cps": [{
                    "cpd_id": 'CP2',
                    "cp_config": [{
                        "link_port_id": uuidsentinel.link_port_id,
                        "cp_protocol_data": [{
                            "layer_protocol": "IP_OVER_ETHERNET"}]}]
                }],
                "ext_link_ports": [{
                    "id": uuidsentinel.link_port_id,
                    "resource_handle": {
                        "resource_id": 'neutron-port-uuid_CP2'}
                }]}]

        ext_mg_vl = [{'id': 'VL3', 'vnf_virtual_link_desc_id': 'VL3',
                      'resource_id': 'neutron-network-uuid_VL3'}]
        request = {'ext_managed_virtual_links': ext_mg_vl,
                   'ext_virtual_links': data, 'flavour_id': 'simple'}
        ctxt = context.get_admin_context()
        inst_req_info = objects.InstantiateVnfRequest.obj_from_primitive(
            request, ctxt)

        # image and info
        grant_info = {
            'VDU1': [objects.VnfResource(id=uuidsentinel.id,
                    vnf_instance_id=uuidsentinel.vnf_instance_id,
                    resource_type='image',
                    resource_identifier='glance-image-uuid_VDU1',
                    tenant_id=uuidsentinel.tenant_id)]}

        self.tth._generate_hot_from_tosca(vnfd_dict, dev_attrs,
                                     inst_req_info, grant_info)

        expected_hot_tpl = self._load_yaml(hot_file)
        actual_hot_tpl = yaml.safe_load(self.tth.heat_template_yaml)
        self.assertEqual(expected_hot_tpl, actual_hot_tpl)

    def test_generate_hot_from_tosca_with_scaling(self):
        tosca_file = './data/etsi_nfv/' \
            'tosca_generate_hot_from_tosca_with_scaling.yaml'
        hot_file = './data/etsi_nfv/hot/' \
            'scaling/' \
            'hot_generate_hot_from_tosca_with_scaling.yaml'
        hot_aspect_file = './data/etsi_nfv/hot/' \
            'scaling/' \
            'worker_instance.hot.yaml'
        vnfd_dict = self._load_yaml(tosca_file, update_import=True)

        # Input params
        dev_attrs = {}

        data = [{
            "id": 'VL1',
            "resource_id": 'neutron-network-uuid_VL1',
            "ext_cps": [{
                "cpd_id": "CP1",
                "cp_config": [{
                    "cp_protocol_data": [{
                        "layer_protocol": "IP_OVER_ETHERNET",
                        "ip_over_ethernet": {
                            "ip_addresses": [{
                                'type': 'IPV4',
                                'subnet_id': 'neutron-subnet-uuid_CP1'}]}}]
                }]}]},
            {
            "id": 'VL2',
                "resource_id": 'neutron-network-uuid_VL2',
                "ext_cps": [{
                    "cpd_id": 'CP2',
                    "cp_config": [{
                        "link_port_id": uuidsentinel.link_port_id,
                        "cp_protocol_data": [{
                            "layer_protocol": "IP_OVER_ETHERNET"}]}]
                }],
                "ext_link_ports": [{
                    "id": uuidsentinel.link_port_id,
                    "resource_handle": {
                        "resource_id": 'neutron-port-uuid_CP2'}
                }]}]

        ext_mg_vl = [{'id': 'VL3', 'vnf_virtual_link_desc_id': 'VL3',
                      'resource_id': 'neutron-network-uuid_VL3'}]
        request = {'ext_managed_virtual_links': ext_mg_vl,
                   'ext_virtual_links': data, 'flavour_id': 'simple',
                   'instantiation_level_id': 'instantiation_level_1'}
        ctxt = context.get_admin_context()
        inst_req_info = objects.InstantiateVnfRequest.obj_from_primitive(
            request, ctxt)

        # image and info
        grant_info = {
            'VDU1': [objects.VnfResource(id=uuidsentinel.id,
                    vnf_instance_id=uuidsentinel.vnf_instance_id,
                    resource_type='image',
                    resource_identifier='glance-image-uuid_VDU1',
                    tenant_id=uuidsentinel.tenant_id)]}

        self.tth._generate_hot_from_tosca(vnfd_dict, dev_attrs,
                                     inst_req_info, grant_info)

        expected_hot_tpl = self._load_yaml(hot_file)
        actual_hot_tpl = yaml.safe_load(self.tth.heat_template_yaml)
        self.assertEqual(expected_hot_tpl, actual_hot_tpl)

        expected_hot_aspect_tpl = self._load_yaml(hot_aspect_file)
        actual_hot_aspect_tpl = \
            yaml.safe_load(
                self.tth.nested_resources['worker_instance.hot.yaml'])
        self.assertEqual(expected_hot_aspect_tpl, actual_hot_aspect_tpl)

    def test_generate_hot_from_tosca_with_substitution_mappings_error(self):
        tosca_file = './data/etsi_nfv/' \
            'tosca_generate_hot_from_tosca_' \
            'with_substitution_mappings_error.yaml'
        vnfd_dict = self._load_yaml(tosca_file, update_import=True)

        dev_attrs = {}

        self.assertRaises(vnfm.InvalidParamsForSM,
                          self.tth._generate_hot_from_tosca,
                          vnfd_dict,
                          dev_attrs,
                          None,
                          None)

    def test_generate_hot_from_tosca_with_params_error(self):
        tosca_file = './data/etsi_nfv/' \
            'tosca_generate_hot_from_tosca_with_params_error.yaml'
        param_file = './data/etsi_nfv/' \
            'tosca_params_error.yaml'
        vnfd_dict = self._load_yaml(tosca_file, update_import=True)

        param_yaml = self._get_template(param_file)
        dev_attrs = {
            'param_values': param_yaml
        }

        self.assertRaises(vnfm.ParamYAMLNotWellFormed,
                          self.tth._generate_hot_from_tosca,
                          vnfd_dict,
                          dev_attrs,
                          None,
                          None)

    def test_generate_hot_from_tosca_parser_error(self):
        tosca_file = './data/etsi_nfv/' \
            'tosca_generate_hot_from_tosca_parser_error.yaml'
        vnfd_dict = self._load_yaml(tosca_file, update_import=True)

        # Input params
        dev_attrs = {}

        self.assertRaises(vnfm.ToscaParserFailed,
                          self.tth._generate_hot_from_tosca,
                          vnfd_dict,
                          dev_attrs,
                          None,
                          None)

    def test_generate_hot_from_tosca_translator_error(self):
        tosca_file = './data/etsi_nfv/' \
            'tosca_generate_hot_from_tosca_translator_error.yaml'
        vnfd_dict = self._load_yaml(tosca_file, update_import=True)

        # Input params
        dev_attrs = {}

        self.assertRaises(vnfm.HeatTranslatorFailed,
                          self.tth._generate_hot_from_tosca,
                          vnfd_dict,
                          dev_attrs,
                          None,
                          None)

    def test_generate_hot_from_tosca_with_scaling_invalid_inst_req(self):
        tosca_file = './data/etsi_nfv/' \
            'tosca_generate_hot_from_tosca_with_scaling_invalid_inst_req.yaml'
        vnfd_dict = self._load_yaml(tosca_file, update_import=True)

        # Input params
        dev_attrs = {}

        data = [{
            "id": 'VL1',
            "resource_id": 'neutron-network-uuid_VL1',
            "ext_cps": [{
                "cpd_id": "CP1",
                "cp_config": [{
                    "cp_protocol_data": [{
                        "layer_protocol": "IP_OVER_ETHERNET",
                        "ip_over_ethernet": {
                            "ip_addresses": [{
                                'type': 'IPV4',
                                'fixed_addresses': ['1.1.1.1'],
                                'subnet_id': 'neutron-subnet-uuid_CP1'}]}
                    }]
                }]
            }]}
        ]

        request = {'ext_virtual_links': data, 'flavour_id': 'simple'}
        ctxt = context.get_admin_context()
        inst_req_info = objects.InstantiateVnfRequest.obj_from_primitive(
            request, ctxt)

        self.assertRaises(vnfm.InvalidInstReqInfoForScaling,
                          self.tth._generate_hot_from_tosca,
                          vnfd_dict,
                          dev_attrs,
                          inst_req_info,
                          None)



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""remove service related

Revision ID: 2774a42c7163
Revises: 12a57080b278
Create Date: 2015-11-26 15:47:51.161749

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '2774a42c7163'
down_revision = '12a57080b278'

from alembic import op


def upgrade(active_plugins=None, options=None):
    # commands auto generated by Alembic - please adjust! #
    op.drop_table('servicecontexts')
    op.drop_table('deviceservicecontexts')
    op.drop_table('servicedevicebindings')
    op.drop_table('serviceinstances')
    # end Alembic commands #



# Copyright 2020 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add scale column

Revision ID: ee98bbc0789d
Revises: c47a733f425a
Create Date: 2020-09-11 16:39:04.039173

"""
# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = 'ee98bbc0789d'
down_revision = 'c47a733f425a'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.add_column('vnf_instantiated_info',
                  sa.Column('scale_status', sa.JSON(), nullable=True))



# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from tacker.api.common import attribute_filter
from tacker.common import exceptions as exception


class BaseViewBuilder(object):

    @classmethod
    def validate_filter(cls, filters=None):
        if not filters:
            return

        return attribute_filter.parse_filter_rule(filters,
            target=cls.FLATTEN_ATTRIBUTES)

    @classmethod
    def validate_attribute_fields(cls, all_fields=None, fields=None,
            exclude_fields=None, exclude_default=None):

        if all_fields and (fields or exclude_fields or exclude_default):
            msg = ("Invalid query parameter combination: 'all_fields' "
                   "cannot be combined with 'fields' or 'exclude_fields' "
                   "or 'exclude_default'")
            raise exception.ValidationError(msg)

        if fields and (all_fields or exclude_fields):
            msg = ("Invalid query parameter combination: 'fields' "
                   "cannot be combined with 'all_fields' or 'exclude_fields' ")
            raise exception.ValidationError(msg)

        if exclude_fields and (all_fields or fields or exclude_default):
            msg = ("Invalid query parameter combination: 'exclude_fields' "
                   "cannot be combined with 'all_fields' or 'fields' "
                   "or 'exclude_default'")
            raise exception.ValidationError(msg)

        if exclude_default and (all_fields or exclude_fields):
            msg = ("Invalid query parameter combination: 'exclude_default' "
                   "cannot be combined with 'all_fields' or 'exclude_fields' ")
            raise exception.ValidationError(msg)

        def _validate_complex_attributes(query_parameter, fields):
            msg = ("Invalid query parameter '%(query_parameter)s'. "
                   "Value: %(field)s")
            for field in fields:
                if field in cls.COMPLEX_ATTRIBUTES:
                    continue
                elif '*' in field:
                    # Field should never contain '*' as it's reserved for
                    # special purpose for handling key-value pairs.
                    raise exception.ValidationError(msg %
                        {"query_parameter": query_parameter,
                         "field": field})
                elif field not in cls.FLATTEN_COMPLEX_ATTRIBUTES:
                    # Special case for field with key-value pairs.
                    # In this particular case, key will act as an attribute
                    # in structure so you need to treat it differently than
                    # other fields. All key-value pair field will be post-fix
                    # with '*' in FLATTEN_COMPLEX_ATTRIBUTES. Request
                    # with field which contains '*' will be treated as an
                    # error.
                    special_field = False
                    for attribute in cls.FLATTEN_COMPLEX_ATTRIBUTES:
                        if '*' in attribute and field.startswith(
                                attribute.split('*')[0]):
                            special_field = True

                    if not special_field:
                        raise exception.ValidationError(msg %
                            {"query_parameter": query_parameter,
                             "field": field})

        if fields:
            _validate_complex_attributes("fields", fields.split(','))
        elif exclude_fields:
            _validate_complex_attributes("exclude_fields",
                    exclude_fields.split(","))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import timeutils
from oslo_versionedobjects import base as ovoo_base
from sqlalchemy import exc
from sqlalchemy.orm import joinedload

from tacker.common import exceptions
from tacker.common import utils
from tacker.db import api as db_api
from tacker.db.db_sqlalchemy import api
from tacker.db.db_sqlalchemy import models
from tacker import objects
from tacker.objects import base
from tacker.objects import common
from tacker.objects import fields


LOG = logging.getLogger(__name__)


@db_api.context_manager.writer
def _vnf_lcm_op_occ_create(context, values):
    fields = {
        'id': values.id,
        'operation_state': values.operation_state,
        'state_entered_time': values.state_entered_time,
        'start_time': values.start_time,
        'vnf_instance_id': values.vnf_instance_id,
        'operation': values.operation,
        'is_automatic_invocation': values.is_automatic_invocation,
        'is_cancel_pending': values.is_cancel_pending,
        'error': values.error,
        'resource_changes': values.resource_changes,
        'changed_info': values.changed_info,
        'changed_ext_connectivity': values.changed_ext_connectivity,
        'error_point': values.error_point,
        'tenant_id': values.tenant_id,
    }
    if 'grant_id' in values:
        fields['grant_id'] = values.grant_id
    if 'operation_params' in values:
        fields['operation_params'] = values.operation_params

    context.session.execute(
        models.VnfLcmOpOccs.__table__.insert(None),
        fields)


@db_api.context_manager.writer
def _vnf_lcm_op_occ_update(context, values):
    update = {'operation_state': values.operation_state,
              'state_entered_time': values.state_entered_time,
              'error_point': values.error_point,
              'updated_at': datetime.utcnow()}
    LOG.debug('values %s', values)
    if 'resource_changes' in values:
        if values.resource_changes:
            update.update({'resource_changes': jsonutils.dumps(
                values.resource_changes.to_dict())})
    if 'error' in values:
        if values.error:
            update.update({'error': jsonutils.dumps(values.error.to_dict())})
    if 'changed_info' in values:
        if values.changed_info:
            update.update({'changed_info': jsonutils.dumps(
                values.changed_info.to_dict())})
    if 'changed_ext_connectivity' in values:
        if values.changed_ext_connectivity:
            update.update({'changed_ext_connectivity': jsonutils.dumps(
                [chg_ext_conn.to_dict() for chg_ext_conn in
                values.changed_ext_connectivity])})
    api.model_query(context, models.VnfLcmOpOccs). \
        filter_by(id=values.id). \
        update(update, synchronize_session=False)


def _make_vnf_lcm_op_occs_list(context, op_occ_list,
                               db_op_occ_list):
    lcm_op_occ_class = VnfLcmOpOcc

    op_occ_list.objects = []
    for db_op_occ in db_op_occ_list:
        if(db_op_occ['changed_info'] and
                isinstance(db_op_occ['changed_info'], str)):
            db_op_occ['changed_info'] = jsonutils.loads(
                db_op_occ['changed_info'])
        vnf_lcm_op_occ_obj = lcm_op_occ_class._from_db_object(
            context, lcm_op_occ_class(context), db_op_occ)
        op_occ_list.objects.append(vnf_lcm_op_occ_obj)

    op_occ_list.obj_reset_changes()
    return op_occ_list


@db_api.context_manager.reader
def _vnf_lcm_op_occs_get_by_id(context, vnf_lcm_op_occ_id):

    query = api.model_query(context, models.VnfLcmOpOccs,
                            read_deleted="no", project_only=True). \
        filter_by(id=vnf_lcm_op_occ_id)

    result = query.first()

    if not result:
        raise exceptions.NotFound(resource='table',
                                  name='vnf_lcm_op_occs')

    return result


@db_api.context_manager.reader
def _vnf_lcm_op_occs_get_by_vnf_instance_id(context, vnf_instance_id):

    query = api.model_query(context, models.VnfLcmOpOccs,
                            read_deleted="no", project_only=True). \
        filter_by(vnf_instance_id=vnf_instance_id)

    result = query.first()

    if not result:
        raise exceptions.VnfInstanceNotFound(id=vnf_instance_id)

    return result


@db_api.context_manager.reader
def _vnf_lcm_op_occs_get_by_filters(context, read_deleted=None,
                                    filters=None):

    query = api.model_query(context, models.VnfLcmOpOccs,
                            read_deleted=read_deleted, project_only=True)

    if filters:
        query = common.apply_filters(query, filters)

    return query.all()


@db_api.context_manager.reader
def _vnf_notify_get_by_id(context, vnf_instance_id, columns_to_join=None):

    query = api.model_query(context, models.VnfLcmOpOccs,
                            read_deleted="no", project_only=True). \
        filter_by(id=vnf_instance_id)

    if columns_to_join:
        for column in columns_to_join:
            query = query.options(joinedload(column))

    result = query.first()

    if not result:
        raise exceptions.VnfInstanceNotFound(id=vnf_instance_id)

    return result


@db_api.context_manager.writer
def _vnf_notify_create(context, values):

    vnf_lcm_op_occs = models.VnfLcmOpOccs()
    vnf_lcm_op_occs.update(values)
    vnf_lcm_op_occs.save(context.session)

    return _vnf_notify_get_by_id(context, vnf_lcm_op_occs.id,
                                columns_to_join=None)


@db_api.context_manager.writer
def _vnf_notify_update(context, vnf_instance_id, values,
                      columns_to_join=None):

    vnf_lcm_op_occs = _vnf_notify_get_by_id(context, vnf_instance_id,
                                        columns_to_join=columns_to_join)
    values = values.to_dict()
    vnf_lcm_op_occs.update(values)
    vnf_lcm_op_occs.save(session=context.session)

    return vnf_lcm_op_occs


@db_api.context_manager.writer
def _destroy_vnf_notify(context, uuid):
    now = timeutils.utcnow()
    updated_values = {'deleted': True,
                      'deleted_at': now
                      }
    api.model_query(context, models.VnfLcmOpOccs). \
        filter_by(id=uuid). \
        update(updated_values, synchronize_session=False)


# decorator to catch DBAccess exception
def _wrap_object_error(method):

    def wrapper(*args, **kwargs):
        try:
            method(*args, **kwargs)
        except exc.SQLAlchemyError:
            raise exceptions.DBAccessError

    return wrapper


@base.TackerObjectRegistry.register
class VnfLcmOpOcc(base.TackerObject, base.TackerObjectDictCompat,
                  base.TackerPersistentObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(nullable=False),
        'operation_state': fields.StringField(nullable=False),
        'state_entered_time': fields.DateTimeField(nullable=False),
        'start_time': fields.DateTimeField(nullable=False),
        'vnf_instance_id': fields.StringField(nullable=False),
        'grant_id': fields.StringField(nullable=True),
        'operation': fields.StringField(nullable=False),
        'is_automatic_invocation': fields.BooleanField(default=False),
        'operation_params': fields.StringField(nullable=True),
        'is_cancel_pending': fields.BooleanField(default=False),
        'error': fields.ObjectField(
            'ProblemDetails', nullable=True, default=None),
        'resource_changes': fields.ObjectField(
            'ResourceChanges', nullable=True, default=None),
        'changed_info': fields.ObjectField(
            'VnfInfoModifications', nullable=True, default=None),
        'changed_ext_connectivity': fields.ListOfObjectsField(
            'ExtVirtualLinkInfo', nullable=True, default=[]),
        'error_point': fields.IntegerField(nullable=True, default=0),
        'tenant_id': fields.StringField(nullable=False)
    }

    ALL_ATTRIBUTES = {
        'id': ('id', 'uuid', 'VnfLcmOpOccs'),
        'operationState': ('operation_state', 'string', 'VnfLcmOpOccs'),
        'stateEnteredTime':
            ('state_entered_time', 'datetime', 'VnfLcmOpOccs'),
        'startTime': ('start_time', 'datetime', 'VnfLcmOpOccs'),
        'vnfInstanceId': ('vnf_instance_id', 'string', 'VnfLcmOpOccs'),
        'grantId': ('grant_id', 'string', 'VnfLcmOpOccs'),
        'operation': ('operation', 'string', 'VnfLcmOpOccs'),
        'isAutomaticInvocation':
            ('is_automatic_invocation', 'boolean', 'VnfLcmOpOccs'),
        'isCancelPending': ('is_cancel_pending', 'string', 'VnfLcmOpOccs'),
        'errorPoint': ('error_point', 'number', 'VnfLcmOpOccs'),
        'operationParams': ('operation_params', 'string', 'VnfLcmOpOccs'),
        'error': ('error', 'string', 'VnfLcmOpOccs'),
        'resourceChanges': ('resource_changes', 'string', 'VnfLcmOpOccs'),
        'changedInfo': ('changed_info', 'string', 'VnfLcmOpOccs')
    }

    FLATTEN_ATTRIBUTES = utils.flatten_dict(ALL_ATTRIBUTES.copy())

    SIMPLE_ATTRIBUTES = ['id', 'operationState', 'stateEnteredTime',
                         'startTime', 'vnfInstanceId', 'grantId', 'operation',
                         'isAutomaticInvocation',
                         'isCancelPending', 'errorPoint']

    COMPLEX_ATTRIBUTES = ['error', 'resourceChanges', 'changedInfo',
                          'operationParams', 'changedExtConnectivity']

    @base.remotable
    def create(self):
        updates = self.obj_clone()
        _vnf_lcm_op_occ_create(self._context, updates)

    @base.remotable
    def save(self):
        updates = self.obj_clone()
        _vnf_lcm_op_occ_update(self._context, updates)

    @staticmethod
    def _from_db_object(context, vnf_lcm_op_occ_obj, db_vnf_lcm_op_occ):

        special_fields = ['error',
                          'resource_changes',
                          'changed_info',
                          'changed_ext_connectivity']
        for key in vnf_lcm_op_occ_obj.fields:
            if key in special_fields:
                continue
            setattr(vnf_lcm_op_occ_obj, key, db_vnf_lcm_op_occ.get(key))
        if db_vnf_lcm_op_occ['error']:
            error = ProblemDetails.obj_from_primitive(
                db_vnf_lcm_op_occ['error'], context)
            vnf_lcm_op_occ_obj.error = error
        if db_vnf_lcm_op_occ['resource_changes']:
            resource_changes = ResourceChanges.obj_from_primitive(
                db_vnf_lcm_op_occ['resource_changes'], context)
            vnf_lcm_op_occ_obj.resource_changes = resource_changes
        if db_vnf_lcm_op_occ['changed_info']:
            changed_info = VnfInfoModifications.obj_from_primitive(
                db_vnf_lcm_op_occ['changed_info'], context)
            vnf_lcm_op_occ_obj.changed_info = changed_info
        if db_vnf_lcm_op_occ['changed_ext_connectivity']:
            changed_ext_conn = \
                [objects.ExtVirtualLinkInfo.obj_from_primitive(
                 chg_ext_conn, context) for chg_ext_conn in
                 jsonutils.loads(
                    db_vnf_lcm_op_occ['changed_ext_connectivity'])]
            vnf_lcm_op_occ_obj.changed_ext_connectivity = changed_ext_conn

        vnf_lcm_op_occ_obj._context = context
        vnf_lcm_op_occ_obj.obj_reset_changes()
        return vnf_lcm_op_occ_obj

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            vnf_lcm_op_occ = super(
                VnfLcmOpOcc, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'error' in primitive.keys():
                obj_data = ProblemDetails._from_dict(
                    primitive.get('error'))
                primitive.update({'error': obj_data})
            if 'resource_changes' in primitive.keys():
                obj_data = ResourceChanges._from_dict(
                    primitive.get('resource_changes'))
                primitive.update({'resource_changes': obj_data})
            if 'changed_info' in primitive.keys():
                obj_data = VnfInfoModifications._from_dict(
                    primitive.get('changed_info'))
                primitive.update({'changed_info': obj_data})
            if 'changed_ext_connectivity' in primitive.keys():
                obj_data = [objects.ExtVirtualLinkInfo.obj_from_primitive(
                    chg_ext_conn, context) for chg_ext_conn in
                    primitive.get('changed_ext_connectivity')]
                primitive.update({'changed_ext_connectivity': obj_data})
            vnf_lcm_op_occ = VnfLcmOpOcc._from_dict(primitive)

        return vnf_lcm_op_occ

    @classmethod
    def obj_from_db_obj(cls, context, db_obj):
        return cls._from_db_object(context, cls(), db_obj)

    @classmethod
    def _from_dict(cls, data_dict):
        operation_state = data_dict.get('operation_state')
        state_entered_time = data_dict.get('state_entered_time')
        start_time = data_dict.get('start_time')
        vnf_instance_id = data_dict.get('vnf_instance_id')
        grant_id = data_dict.get('grant_id')
        operation = data_dict.get('operation')
        is_automatic_invocation = data_dict.get('is_automatic_invocation')
        operation_params = data_dict.get('operation_params')
        is_cancel_pending = data_dict.get('is_cancel_pending')
        error = data_dict.get('error')
        resource_changes = data_dict.get('resource_changes')
        changed_info = data_dict.get('changed_info')
        changed_ext_connectivity = data_dict.get('changed_ext_connectivity')
        error_point = data_dict.get('error_point')
        tenant_id = data_dict.get('tenant_id')

        obj = cls(operation_state=operation_state,
                  state_entered_time=state_entered_time,
                  start_time=start_time,
                  vnf_instance_id=vnf_instance_id,
                  grant_id=grant_id,
                  operation=operation,
                  is_automatic_invocation=is_automatic_invocation,
                  operation_params=operation_params,
                  is_cancel_pending=is_cancel_pending,
                  error=error,
                  resource_changes=resource_changes,
                  changed_info=changed_info,
                  changed_ext_connectivity=changed_ext_connectivity,
                  error_point=error_point,
                  tenant_id=tenant_id
                  )

        return obj

    def _get_error(self, include_fields=None):
        key = 'error'
        if key in include_fields:
            return {key: self.error.to_dict()}

    def _get_resource_changes(self, include_fields=None):
        key = 'resourceChanges'
        if key in include_fields:
            return {key: self.resource_changes.to_dict()}

    def _get_changed_info(self, include_fields=None):
        key = 'changedInfo'
        if key in include_fields:
            return {key: self.changed_info.to_dict()}

    def _get_operation_params(self, include_fields=None):
        key = 'operationParams'
        if key in include_fields:
            return {key: self.operation_params}

    def _get_changed_ext_connectivity(self, include_fields=None):
        key = 'changedExtConnectivity'
        return {key: [chg_ext_conn.to_dict() for chg_ext_conn in
                      self.changed_ext_connectivity]}

    def to_dict(self, include_fields=None):
        data = dict()
        if not include_fields:
            include_fields = set(self.FLATTEN_ATTRIBUTES.keys())

        # add simple fields
        to_fields = set(self.SIMPLE_ATTRIBUTES).intersection(include_fields)
        for field in to_fields:
            data[field] = getattr(self, self.FLATTEN_ATTRIBUTES[field][0])

        # add complex attributes
        if self.error:
            error = self._get_error(include_fields=include_fields)
            if error:
                data.update(error)

        if self.resource_changes:
            resource_changes = self._get_resource_changes(
                include_fields=include_fields)
            if resource_changes:
                data.update(resource_changes)

        if self.changed_info:
            changed_info = self._get_changed_info(
                include_fields=include_fields)
            if changed_info:
                data.update(changed_info)

        if self.operation_params:
            operation_params = self._get_operation_params(
                include_fields=include_fields)
            if operation_params:
                data.update(operation_params)

        if self.changed_ext_connectivity:
            changed_ext_connectivity = self._get_changed_ext_connectivity(
                include_fields=include_fields)
            if changed_ext_connectivity:
                data.update(changed_ext_connectivity)

        return data

    @base.remotable_classmethod
    def get_by_id(cls, context, id):
        db_vnf_lcm_op_occs = _vnf_lcm_op_occs_get_by_id(context, id)
        return cls._from_db_object(context, cls(), db_vnf_lcm_op_occs)

    @base.remotable_classmethod
    def get_by_vnf_instance_id(cls, context, id):
        db_vnf_lcm_op_occs = _vnf_lcm_op_occs_get_by_vnf_instance_id(
            context, id)
        return cls._from_db_object(context, cls(), db_vnf_lcm_op_occs)


@base.TackerObjectRegistry.register
class VnfLcmOpOccList(ovoo_base.ObjectListBase, base.TackerObject):

    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('VnfLcmOpOcc')
    }

    @base.remotable_classmethod
    def get_by_filters(cls, context, read_deleted=None, filters=None):
        db_vnf_lcm_op_occs = _vnf_lcm_op_occs_get_by_filters(
            context, read_deleted=read_deleted, filters=filters)
        return _make_vnf_lcm_op_occs_list(context, cls(), db_vnf_lcm_op_occs)


@base.TackerObjectRegistry.register
class ResourceChanges(base.TackerObject,
                     base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'affected_vnfcs': fields.ListOfObjectsField(
            'AffectedVnfc', nullable=True),
        'affected_virtual_links': fields.ListOfObjectsField(
            'AffectedVirtualLink', nullable=True),
        'affected_virtual_storages': fields.ListOfObjectsField(
            'AffectedVirtualStorage', nullable=True)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            resource_changes = super(
                ResourceChanges, cls).obj_from_primitive(
                primitive, context)
        else:
            rs_dict = jsonutils.loads(primitive)
            if rs_dict.get('affected_vnfcs'):
                obj_data = [AffectedVnfc._from_dict(
                    affected_vnfc) for affected_vnfc in rs_dict.get(
                    'affected_vnfcs', [])]
                rs_dict.update({'affected_vnfcs': obj_data})
            if rs_dict.get('affected_virtual_links'):
                obj_data = [AffectedVirtualLink._from_dict(
                    affected_virtual_link)
                    for affected_virtual_link in rs_dict.get(
                    'affected_virtual_links', [])]
                rs_dict.update({'affected_virtual_links': obj_data})
            if rs_dict.get('affected_virtual_storages'):
                obj_data = [AffectedVirtualStorage._from_dict(
                    affected_virtual_storage)
                    for affected_virtual_storage in rs_dict.get(
                    'affected_virtual_storages', [])]
                rs_dict.update({'affected_virtual_storages': obj_data})
            resource_changes = ResourceChanges._from_dict(rs_dict)

        return resource_changes

    @classmethod
    def _from_dict(cls, data_dict):
        affected_vnfcs = data_dict.get('affected_vnfcs')
        affected_virtual_links = data_dict.get('affected_virtual_links')
        affected_virtual_storages = data_dict.get('affected_virtual_storages')

        obj = cls(affected_vnfcs=affected_vnfcs,
                  affected_virtual_links=affected_virtual_links,
                  affected_virtual_storages=affected_virtual_storages
                  )

        return obj

    def to_dict(self):
        data = {}
        if self.affected_vnfcs:
            affected_vnfcs_list = []
            for affected_vnfc in self.affected_vnfcs:
                affected_vnfcs_list.append(affected_vnfc.to_dict())

            data.update({'affected_vnfcs': affected_vnfcs_list})
        if self.affected_virtual_links:
            affected_virtual_links_list = []
            for affected_virtual_link in self.affected_virtual_links:
                affected_virtual_links_list.append(
                    affected_virtual_link.to_dict())

            data.update(
                {'affected_virtual_links': affected_virtual_links_list})
        if self.affected_virtual_storages:
            affected_virtual_storages_list = []
            for affected_virtual_storage in self.affected_virtual_storages:
                affected_virtual_storages_list.append(
                    affected_virtual_storage.to_dict())

            data.update(
                {'affected_virtual_storages': affected_virtual_storages_list})
        return data


@base.TackerObjectRegistry.register
class ProblemDetails(base.TackerObject,
                     base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'title': fields.StringField(nullable=True, default=''),
        'status': fields.IntegerField(nullable=False),
        'detail': fields.StringField(nullable=False)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            problem_detail = super(
                ProblemDetails, cls).obj_from_primitive(
                primitive, context)
        else:
            p_dict = jsonutils.loads(primitive)
            problem_detail = ProblemDetails._from_dict(p_dict)

        return problem_detail

    @classmethod
    def _from_dict(cls, data_dict):
        title = data_dict.get('title')
        status = data_dict.get('status')
        detail = data_dict.get('detail')

        obj = cls(title=title,
                  status=status,
                  detail=detail)

        return obj

    def to_dict(self):
        return {'title': self.title,
                'status': self.status,
                'detail': self.detail}


@base.TackerObjectRegistry.register
class AffectedVnfc(base.TackerObject,
                   base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'vdu_id': fields.StringField(nullable=False),
        'change_type': fields.StringField(nullable=False),
        'compute_resource': fields.ObjectField(
            'ResourceHandle', nullable=False),
        'affected_vnfc_cp_ids':
            fields.ListOfStringsField(nullable=True, default=[]),
        'added_storage_resource_ids':
            fields.ListOfStringsField(nullable=True, default=[]),
        'removed_storage_resource_ids':
            fields.ListOfStringsField(nullable=True, default=[])
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            affected_vnfc = super(
                AffectedVnfc, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'compute_resource' in primitive.keys():
                obj_data = ResourceHandle._from_dict(
                    primitive.get('compute_resource'))
                primitive.update({'compute_resource': obj_data})
            affected_vnfc = AffectedVnfc._from_dict(primitive)

        return affected_vnfc

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        vdu_id = data_dict.get('vdu_id')
        change_type = data_dict.get('change_type')
        compute_resource = ResourceHandle._from_dict(
            data_dict.get('compute_resource'))
        affected_vnfc_cp_ids = data_dict.get('affected_vnfc_cp_ids')
        added_storage_resource_ids = data_dict.get(
            'added_storage_resource_ids')
        removed_storage_resource_ids = data_dict.get(
            'removed_storage_resource_ids')

        obj = cls(id=id,
                  vdu_id=vdu_id,
                  change_type=change_type,
                  compute_resource=compute_resource,
                  affected_vnfc_cp_ids=affected_vnfc_cp_ids,
                  added_storage_resource_ids=added_storage_resource_ids,
                  removed_storage_resource_ids=removed_storage_resource_ids
                  )

        return obj

    def to_dict(self):
        return {
            'id': self.id,
            'vdu_id': self.vdu_id,
            'change_type': self.change_type,
            'compute_resource': self.compute_resource.to_dict(),
            'affected_vnfc_cp_ids': self.affected_vnfc_cp_ids,
            'added_storage_resource_ids': self.added_storage_resource_ids,
            'removed_storage_resource_ids': self.removed_storage_resource_ids}


@base.TackerObjectRegistry.register
class AffectedVirtualLink(base.TackerObject,
                   base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'vnf_virtual_link_desc_id': fields.StringField(nullable=False),
        'change_type': fields.StringField(nullable=False),
        'network_resource': fields.ObjectField(
            'ResourceHandle', nullable=False)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            affected_virtual_link = super(
                AffectedVirtualLink, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'network_resource' in primitive.keys():
                obj_data = ResourceHandle._from_dict(
                    primitive.get('network_resource'))
                primitive.update({'network_resource': obj_data})
            affected_virtual_link = AffectedVirtualLink._from_dict(primitive)

        return affected_virtual_link

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        vnf_virtual_link_desc_id = data_dict.get('vnf_virtual_link_desc_id')
        change_type = data_dict.get('change_type')
        network_resource = ResourceHandle._from_dict(
            data_dict.get('network_resource'))

        obj = cls(id=id,
                  vnf_virtual_link_desc_id=vnf_virtual_link_desc_id,
                  change_type=change_type,
                  network_resource=network_resource
                  )

        return obj

    def to_dict(self):
        return {'id': self.id,
                'vnf_virtual_link_desc_id': self.vnf_virtual_link_desc_id,
                'change_type': self.change_type,
                'network_resource': self.network_resource.to_dict()}


@base.TackerObjectRegistry.register
class AffectedVirtualStorage(base.TackerObject,
                   base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.StringField(nullable=False),
        'virtual_storage_desc_id': fields.StringField(nullable=False),
        'change_type': fields.StringField(nullable=False),
        'storage_resource': fields.ObjectField(
            'ResourceHandle', nullable=False)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            affected_virtual_storage = super(
                AffectedVirtualStorage, cls).obj_from_primitive(
                primitive, context)
        else:
            if 'storage_resource' in primitive.keys():
                obj_data = ResourceHandle._from_dict(
                    primitive.get('storage_resource'))
                primitive.update({'storage_resource': obj_data})
            affected_virtual_storage = AffectedVirtualStorage._from_dict(
                primitive)

        return affected_virtual_storage

    @classmethod
    def _from_dict(cls, data_dict):
        id = data_dict.get('id')
        virtual_storage_desc_id = data_dict.get('virtual_storage_desc_id')
        change_type = data_dict.get('change_type')
        storage_resource = ResourceHandle._from_dict(
            data_dict.get('storage_resource'))

        obj = cls(id=id,
                  virtual_storage_desc_id=virtual_storage_desc_id,
                  change_type=change_type,
                  storage_resource=storage_resource
                  )

        return obj

    def to_dict(self):
        return {'id': self.id,
                'virtual_storage_desc_id': self.virtual_storage_desc_id,
                'change_type': self.change_type,
                'storage_resource': self.storage_resource.to_dict()}


@base.TackerObjectRegistry.register
class VnfInfoModifications(base.TackerObject,
                   base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'vnf_instance_name': fields.StringField(nullable=True),
        'vnf_instance_description': fields.StringField(nullable=True),
        'metadata': fields.DictOfStringsField(nullable=True, default={}),
        'vim_connection_info': fields.ListOfObjectsField(
            'VimConnectionInfo', nullable=True, default=[]),
        'vim_connection_info_delete_ids':
            fields.ListOfStringsField(nullable=True, default=[]),
        'vnf_pkg_id': fields.StringField(nullable=True, default=None),
        'vnfd_id': fields.StringField(nullable=True),
        'vnf_provider': fields.StringField(nullable=True),
        'vnf_product_name': fields.StringField(nullable=True),
        'vnf_software_version': fields.StringField(nullable=True),
        'vnfd_version': fields.StringField(nullable=True)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            vnf_info_modifications = super(
                VnfInfoModifications, cls).obj_from_primitive(
                primitive, context)
        else:
            if isinstance(primitive, str):
                primitive = jsonutils.loads(primitive)

            if 'vim_connection_info' in primitive.keys():
                obj_data = [objects.VimConnectionInfo._from_dict(
                    vim_conn) for vim_conn in primitive.get(
                    'vim_connection_info', [])]
                primitive.update({'vim_connection_info': obj_data})
            vnf_info_modifications = VnfInfoModifications._from_dict(primitive)

        return vnf_info_modifications

    @classmethod
    def _from_dict(cls, data_dict):
        vnf_instance_name = data_dict.get('vnf_instance_name')
        vnf_instance_description = data_dict.get('vnf_instance_description')
        metadata = data_dict.get('metadata')
        vim_connection_info = data_dict.get('vim_connection_info', [])
        vim_connection_info_delete_ids = data_dict.get(
            'vim_connection_info_delete_ids')
        vnf_pkg_id = data_dict.get('vnf_pkg_id')
        vnfd_id = data_dict.get('vnfd_id')
        vnf_provider = data_dict.get('vnf_provider')
        vnf_product_name = data_dict.get('vnf_product_name')
        vnf_software_version = data_dict.get('vnf_software_version')
        vnfd_version = data_dict.get('vnfd_version')

        obj = cls(
            vnf_instance_name=vnf_instance_name,
            vnf_instance_description=vnf_instance_description,
            metadata=metadata,
            vim_connection_info=vim_connection_info,
            vim_connection_info_delete_ids=vim_connection_info_delete_ids,
            vnf_pkg_id=vnf_pkg_id,
            vnfd_id=vnfd_id,
            vnf_provider=vnf_provider,
            vnf_product_name=vnf_product_name,
            vnf_software_version=vnf_software_version,
            vnfd_version=vnfd_version)

        return obj

    def to_dict(self):
        """For the attributes of this class, if an attribute exists and is not

        null, it means that the attribute has been modified. This method

        returns a dictionary containing the modified attributes.
        """
        dct = {}
        for field in self.fields:
            if field in self and getattr(self, field):
                value = getattr(self, field)
                # Since the type of vim_connection_info is ListOfObjectsField,
                # the objects in vim_connection_info also need to be converted
                # into dictionary, otherwise an error will occur
                # when serialized.
                if field == "vim_connection_info":
                    value = [vim_conn.to_dict() for vim_conn in value]
                dct[field] = value
        return dct


@base.TackerObjectRegistry.register
class ResourceHandle(base.TackerObject,
                     base.TackerPersistentObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'vim_connection_id': fields.StringField(nullable=True,
                                                default=None),
        'resource_id': fields.StringField(nullable=False, default=""),
        'vim_level_resource_type': fields.StringField(nullable=True,
                                                      default=None)
    }

    @classmethod
    def obj_from_primitive(cls, primitive, context):
        if 'tacker_object.name' in primitive:
            resource_handle = super(
                ResourceHandle, cls).obj_from_primitive(
                primitive, context)
        else:
            resource_handle = ResourceHandle._from_dict(primitive)

        return resource_handle

    @classmethod
    def _from_dict(cls, data_dict):
        LOG.debug("data_dict %s", data_dict)
        vim_connection_id = data_dict.get('vim_connection_id')
        resource_id = data_dict.get('resource_id', "")
        vim_level_resource_type = data_dict.get('vim_level_resource_type')

        obj = cls(vim_connection_id=vim_connection_id,
                  resource_id=resource_id,
                  vim_level_resource_type=vim_level_resource_type)

        return obj

    def to_dict(self):
        return {'vim_connection_id': self.vim_connection_id,
                'resource_id': self.resource_id,
                'vim_level_resource_type': self.vim_level_resource_type}



# Copyright 2019 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures

from heatclient import client
from keystoneauth1 import fixture
from keystoneauth1 import loading
from keystoneauth1 import session
from openstack import connection

IDENTITY_URL = 'http://identityserver:5000/v3'
HEAT_URL = 'http://heat-api'
GLANCE_URL = 'http://image-api/v2'


class ClientFixture(fixtures.Fixture):

    def __init__(self, requests_mock, heat_url=HEAT_URL,
                 identity_url=IDENTITY_URL):
        super(ClientFixture, self).__init__()
        self.identity_url = identity_url
        self.client = None
        self.token = fixture.V2Token()
        self.token.set_scope()
        self.requests_mock = requests_mock
        self.discovery = fixture.V2Discovery(href=self.identity_url)
        s = self.token.add_service('orchestration')
        s.add_endpoint(heat_url)
        self.auth_url = '%s/tokens' % self.identity_url

    def setUp(self):
        super(ClientFixture, self).setUp()
        headers = {'X-Content-Type': 'application/json'}
        self.requests_mock.post(self.auth_url,
                      json=self.token, headers=headers)
        self.requests_mock.get(self.identity_url,
                      json=self.discovery, headers=headers)
        self.client = self.new_client()

    def _set_session(self):
        self.session = session.Session()
        loader = loading.get_plugin_loader('password')
        self.session.auth = loader.load_from_options(
            auth_url=self.identity_url, username='xx', password='xx')

    def new_client(self):
        self._set_session()
        return client.Client("1", session=self.session)


class SdkConnectionFixture(ClientFixture):
    """Fixture class to access the apis via openstacksdk's Connection object.

        This class is mocking the requests of glance api.
    """

    def __init__(self, requests_mock, glance_url=GLANCE_URL):
        super(SdkConnectionFixture, self).__init__(requests_mock)
        s = self.token.add_service('image')
        s.add_endpoint(glance_url)

    def new_client(self):
        self._set_session()
        conn = connection.Connection(
            region_name=None,
            session=self.session,
            identity_interface='internal',
            image_api_version='2')
        return conn



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from tacker.common import log
from tacker import context as tacker_context
from tacker.sol_refactored.common import config
from tacker.sol_refactored.common import coordinate
from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.sol_refactored.common import lcm_op_occ_utils as lcmocc_utils
from tacker.sol_refactored.common import vnf_instance_utils as inst_utils
from tacker.sol_refactored.conductor import vnflcm_driver_v2
from tacker.sol_refactored.nfvo import nfvo_client
from tacker.sol_refactored import objects
from tacker.sol_refactored.objects.v2 import fields


LOG = logging.getLogger(__name__)

CONF = config.CONF


class ConductorV2(object):

    def __init__(self):
        self.vnflcm_driver = vnflcm_driver_v2.VnfLcmDriverV2()
        self.endpoint = CONF.v2_vnfm.endpoint
        self.nfvo_client = nfvo_client.NfvoClient()

        self._change_lcm_op_state()

    def _change_lcm_op_state(self):
        # NOTE: If the conductor down during processing and
        # the LcmOperationState STARTING/PROCESSING/ROLLING_BACK remain,
        # change it at the next startup.
        context = tacker_context.get_admin_context()
        ex = sol_ex.ConductorProcessingError()

        state_list = [(fields.LcmOperationStateType.STARTING,
                       fields.LcmOperationStateType.ROLLED_BACK),
                      (fields.LcmOperationStateType.PROCESSING,
                       fields.LcmOperationStateType.FAILED_TEMP),
                      (fields.LcmOperationStateType.ROLLING_BACK,
                       fields.LcmOperationStateType.FAILED_TEMP)]
        for before_state, after_state in state_list:
            lcmoccs = objects.VnfLcmOpOccV2.get_by_filter(context,
                operationState=before_state)
            for lcmocc in lcmoccs:
                lcmocc.operationState = after_state
                self._set_lcmocc_error(lcmocc, ex)
                inst = inst_utils.get_inst(context, lcmocc.vnfInstanceId)
                lcmocc.update(context)
                # send notification
                self.nfvo_client.send_lcmocc_notification(context, lcmocc,
                                                          inst, self.endpoint)

    def _set_lcmocc_error(self, lcmocc, ex):
        if isinstance(ex, sol_ex.SolException):
            problem_details = ex.make_problem_details()
        else:
            # program bug. it occurs only under development.
            problem_details = {'status': 500,
                               'detail': str(ex)}
        lcmocc.error = objects.ProblemDetails.from_dict(problem_details)

    @log.log
    def start_lcm_op(self, context, lcmocc_id):
        lcmocc = lcmocc_utils.get_lcmocc(context, lcmocc_id)

        self._start_lcm_op(context, lcmocc)

    @coordinate.lock_vnf_instance('{lcmocc.vnfInstanceId}', delay=True)
    def _start_lcm_op(self, context, lcmocc):
        # just consistency check
        if lcmocc.operationState != fields.LcmOperationStateType.STARTING:
            LOG.error("VnfLcmOpOcc unexpected operationState.")
            return

        inst = inst_utils.get_inst(context, lcmocc.vnfInstanceId)

        # NOTE: error cannot happen to here basically.
        # if an error occurred lcmocc.opetationState remains STARTING.
        # see the log of the tacker-conductor to investigate the cause
        # of error.

        # NOTE: the following flow follows SOL003 5.4.1.2

        # send notification STARTING
        self.nfvo_client.send_lcmocc_notification(context, lcmocc, inst,
                                                  self.endpoint)

        try:
            vnfd = self.nfvo_client.get_vnfd(context, inst.vnfdId,
                                             all_contents=True)

            # NOTE: perform grant exchange mainly but also perform
            # something to do at STATING phase ex. request check.
            grant_req, grant = self.vnflcm_driver.grant(context, lcmocc,
                                                        inst, vnfd)
            self.vnflcm_driver.post_grant(context, lcmocc, inst, grant_req,
                                          grant, vnfd)

            lcmocc.operationState = fields.LcmOperationStateType.PROCESSING
            lcmocc.grantId = grant.id
            with context.session.begin(subtransactions=True):
                # save grant_req and grant to be used when retry
                # NOTE: grant_req is saved because it is necessary to interpret
                # the contents of grant. Though grant can be gotten from NFVO,
                # it is saved here with grant_req so that it is not necessary
                # to communicate with NFVO when retry. They are saved temporary
                # and will be deleted when operationState becomes an end state
                # (COMPLETED/FAILED/ROLLED_BACK).
                grant_req.create(context)
                grant.create(context)
                lcmocc.update(context)
        except Exception as ex:
            LOG.exception("STARTING %s failed", lcmocc.operation)
            lcmocc.operationState = fields.LcmOperationStateType.ROLLED_BACK
            self._set_lcmocc_error(lcmocc, ex)
            lcmocc.update(context)

        # send notification PROCESSING or ROLLED_BACK
        self.nfvo_client.send_lcmocc_notification(context, lcmocc, inst,
                                                  self.endpoint)

        if lcmocc.operationState != fields.LcmOperationStateType.PROCESSING:
            return

        try:
            self.vnflcm_driver.process(context, lcmocc, inst, grant_req,
                                       grant, vnfd)

            lcmocc.operationState = fields.LcmOperationStateType.COMPLETED
            # update inst and lcmocc at the same time
            with context.session.begin(subtransactions=True):
                inst.update(context)
                lcmocc.update(context)
                # grant_req and grant are not necessary any more.
                grant_req.delete(context)
                grant.delete(context)
        except Exception as ex:
            LOG.exception("PROCESSING %s failed", lcmocc.operation)
            lcmocc.operationState = fields.LcmOperationStateType.FAILED_TEMP
            self._set_lcmocc_error(lcmocc, ex)
            lcmocc.update(context)

        # send notification COMPLETED or FAILED_TEMP
        self.nfvo_client.send_lcmocc_notification(context, lcmocc, inst,
                                                  self.endpoint)

    @log.log
    def retry_lcm_op(self, context, lcmocc_id):
        lcmocc = lcmocc_utils.get_lcmocc(context, lcmocc_id)

        self._retry_lcm_op(context, lcmocc)

    @coordinate.lock_vnf_instance('{lcmocc.vnfInstanceId}', delay=True)
    def _retry_lcm_op(self, context, lcmocc):
        # just consistency check
        if lcmocc.operationState != fields.LcmOperationStateType.FAILED_TEMP:
            LOG.error("VnfLcmOpOcc unexpected operationState.")
            return

        inst = inst_utils.get_inst(context, lcmocc.vnfInstanceId)

        lcmocc.operationState = fields.LcmOperationStateType.PROCESSING
        lcmocc.update(context)
        # send notification PROCESSING
        self.nfvo_client.send_lcmocc_notification(context, lcmocc, inst,
                                                  self.endpoint)

        try:
            vnfd = self.nfvo_client.get_vnfd(context, inst.vnfdId,
                                             all_contents=True)
            grant_req, grant = lcmocc_utils.get_grant_req_and_grant(context,
                                                                    lcmocc)
            self.vnflcm_driver.post_grant(context, lcmocc, inst, grant_req,
                                          grant, vnfd)
            self.vnflcm_driver.process(context, lcmocc, inst, grant_req,
                                       grant, vnfd)

            lcmocc.operationState = fields.LcmOperationStateType.COMPLETED
            lcmocc.error = None  # clear error
            # update inst and lcmocc at the same time
            with context.session.begin(subtransactions=True):
                inst.update(context)
                lcmocc.update(context)
                # grant_req and grant are not necessary any more.
                if grant_req is not None:
                    grant_req.delete(context)
                    grant.delete(context)
        except Exception as ex:
            LOG.exception("PROCESSING %s failed", lcmocc.operation)
            lcmocc.operationState = fields.LcmOperationStateType.FAILED_TEMP
            self._set_lcmocc_error(lcmocc, ex)
            lcmocc.update(context)
            # grant_req and grant are already saved. they are not deleted
            # while operationState is FAILED_TEMP.

        # send notification COMPLETED or FAILED_TEMP
        self.nfvo_client.send_lcmocc_notification(context, lcmocc, inst,
                                                  self.endpoint)

    @log.log
    def rollback_lcm_op(self, context, lcmocc_id):
        lcmocc = lcmocc_utils.get_lcmocc(context, lcmocc_id)

        self._rollback_lcm_op(context, lcmocc)

    @coordinate.lock_vnf_instance('{lcmocc.vnfInstanceId}', delay=True)
    def _rollback_lcm_op(self, context, lcmocc):
        # just consistency check
        if lcmocc.operationState != fields.LcmOperationStateType.FAILED_TEMP:
            LOG.error("VnfLcmOpOcc unexpected operationState.")
            return

        inst = inst_utils.get_inst(context, lcmocc.vnfInstanceId)

        lcmocc.operationState = fields.LcmOperationStateType.ROLLING_BACK
        lcmocc.update(context)
        # send notification ROLLING_BACK
        self.nfvo_client.send_lcmocc_notification(context, lcmocc, inst,
                                                  self.endpoint)

        try:
            vnfd = self.nfvo_client.get_vnfd(context, inst.vnfdId)
            grant_req, grant = lcmocc_utils.get_grant_req_and_grant(context,
                                                                    lcmocc)
            self.vnflcm_driver.post_grant(context, lcmocc, inst, grant_req,
                                          grant, vnfd)
            self.vnflcm_driver.rollback(context, lcmocc, inst, grant_req,
                                        grant, vnfd)

            lcmocc.operationState = fields.LcmOperationStateType.ROLLED_BACK
            with context.session.begin(subtransactions=True):
                lcmocc.update(context)
                # NOTE: Basically inst is not changed. But there is a case
                # that VIM resources may be changed while rollback. Only
                # change_ext_conn_rollback at the moment.
                if lcmocc.operation == fields.LcmOperationType.CHANGE_EXT_CONN:
                    inst.update(context)
                # grant_req and grant are not necessary any more.
                if grant_req is not None:
                    grant_req.delete(context)
                    grant.delete(context)
        except Exception as ex:
            LOG.exception("ROLLING_BACK %s failed", lcmocc.operation)
            lcmocc.operationState = fields.LcmOperationStateType.FAILED_TEMP
            self._set_lcmocc_error(lcmocc, ex)
            lcmocc.update(context)
            # grant_req and grant are already saved. they are not deleted
            # while operationState is FAILED_TEMP.

        # send notification ROLLED_BACK or FAILED_TEMP
        self.nfvo_client.send_lcmocc_notification(context, lcmocc, inst,
                                                  self.endpoint)

    @log.log
    def modify_vnfinfo(self, context, lcmocc_id):
        lcmocc = lcmocc_utils.get_lcmocc(context, lcmocc_id)

        self._modify_vnfinfo(context, lcmocc)

    @coordinate.lock_vnf_instance('{lcmocc.vnfInstanceId}', delay=True)
    def _modify_vnfinfo(self, context, lcmocc):
        # just consistency check
        if lcmocc.operationState != fields.LcmOperationStateType.PROCESSING:
            LOG.error("VnfLcmOpOcc unexpected operationState.")
            return

        inst = inst_utils.get_inst(context, lcmocc.vnfInstanceId)
        # send notification PROCESSING
        self.nfvo_client.send_lcmocc_notification(context, lcmocc, inst,
                                                  self.endpoint)

        try:
            vnfd = self.nfvo_client.get_vnfd(context, inst.vnfdId)
            self.vnflcm_driver.process(context, lcmocc, inst, None, None, vnfd)
            lcmocc.operationState = fields.LcmOperationStateType.COMPLETED
            # update inst and lcmocc at the same time
            with context.session.begin(subtransactions=True):
                inst.update(context)
                lcmocc.update(context)
        except Exception as ex:
            LOG.exception("PROCESSING %s failed", lcmocc.operation)
            lcmocc.operationState = fields.LcmOperationStateType.FAILED_TEMP
            self._set_lcmocc_error(lcmocc, ex)
            lcmocc.update(context)

        # send notification COMPLETED or FAILED_TEMP
        self.nfvo_client.send_lcmocc_notification(context, lcmocc, inst,
                                                  self.endpoint)



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.api import views as base
from tacker.objects import vnf_package as _vnf_package


class ViewBuilder(base.BaseViewBuilder):

    FLATTEN_ATTRIBUTES = _vnf_package.VnfPackage.FLATTEN_ATTRIBUTES
    COMPLEX_ATTRIBUTES = _vnf_package.VnfPackage.COMPLEX_ATTRIBUTES
    FLATTEN_COMPLEX_ATTRIBUTES = [key for key in FLATTEN_ATTRIBUTES.keys()
        if '/' in key]

    def _get_links(self, vnf_package):
        return {
            "_links": {
                "self": {
                    "href": '/vnfpkgm/v1/vnf_packages/%s'
                            % vnf_package.id
                },
                "packageContent": {
                    "href": '/vnfpkgm/v1/vnf_packages/%s/package_content'
                            % vnf_package.id
                }
            }
        }

    def _get_vnf_package(self, vnf_package, include_fields=None):
        vnf_package_response = vnf_package.to_dict(
            include_fields=include_fields)

        links = self._get_links(vnf_package)
        vnf_package_response.update(links)

        return vnf_package_response

    def _get_modified_user_data(self, old_user_data, new_user_data):
        # Checking for the new keys
        user_data_response = {k: new_user_data[k] for k
                              in set(new_user_data) - set(old_user_data)}

        # Checking for updation in values of existing keys
        for old_key, old_value in old_user_data.items():
            if old_key in new_user_data.keys() and \
                    new_user_data[old_key] != old_user_data[old_key]:
                user_data_response[old_key] = new_user_data[old_key]

        return user_data_response

    def create(self, vnf_package):
        return self._get_vnf_package(vnf_package)

    def show(self, vnf_package):
        return self._get_vnf_package(vnf_package)

    def patch(self, vnf_package, new_vnf_package):
        response = {}
        if vnf_package.operational_state != new_vnf_package.operational_state:
            response['operationalState'] = new_vnf_package.operational_state
        if vnf_package.user_data != new_vnf_package.user_data:
            updated_user_data = self._get_modified_user_data(
                vnf_package.user_data, new_vnf_package.user_data)
            response['userDefinedData'] = updated_user_data

        return response

    def index(self, vnf_packages, all_fields=True,
            exclude_fields=None, fields=None, exclude_default=False):

        # Find out which fields are to be returned in the response.
        if all_fields:
            include_fields = set(self.FLATTEN_ATTRIBUTES.keys())
        if exclude_default and fields or fields:
            # Note(tpatil): If fields contains any of the complex attributes
            # base name, it should include all attributes from that complex
            # attribute. For example, if softwareImages is passed in the
            # fields, it should include all attributes matching
            # softwareImages/*.
            fields = set(fields.split(','))
            attributes = set(self.COMPLEX_ATTRIBUTES).intersection(fields)
            for attribute in attributes:
                add_fields = set([key for key in self.FLATTEN_ATTRIBUTES.
                    keys() if key.startswith(attribute)])
                fields = fields.union(add_fields)

            include_fields = set(_vnf_package.VnfPackage.simple_attributes +
                _vnf_package.VnfPackage.simple_instantiated_attributes). \
                union(fields)
        elif exclude_default:
            include_fields = set(_vnf_package.VnfPackage.simple_attributes +
                _vnf_package.VnfPackage.simple_instantiated_attributes)
        elif exclude_fields:
            # Note(tpatil): If exclude_fields contains any of the complex
            # attributes base name, it should exclude all attributes from
            # that complex attribute. For example, if softwareImages is passed
            # in the excluded_fields, it should exclude all attributes
            # matching softwareImages/*.
            exclude_fields = set(exclude_fields.split(','))
            exclude_additional_attributes = set(
                self.COMPLEX_ATTRIBUTES).intersection(exclude_fields)
            for attribute in exclude_additional_attributes:
                fields = set([key for key in self.FLATTEN_ATTRIBUTES.keys()
                    if key.startswith(attribute)])
                exclude_fields = exclude_fields.union(fields)

            include_fields = set(self.FLATTEN_ATTRIBUTES.keys()) - \
                exclude_fields
        return [self._get_vnf_package(vnf_package,
            include_fields=include_fields)for vnf_package in vnf_packages]



# Copyright (C) 2020 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from tacker.vnfm.mgmt_drivers import vnflcm_abstract_driver


LOG = logging.getLogger(__name__)


class VnflcmMgmtNoop(vnflcm_abstract_driver.VnflcmMgmtAbstractDriver):
    def get_type(self):
        return 'vnflcm_noop'

    def get_name(self):
        return 'vnflcm_noop'

    def get_description(self):
        return 'Tacker VNFMgmt VnflcmNoop Driver'

    def instantiate_start(self, vnf_instance, additional_params, **kwargs):
        LOG.debug('instantiate_start %(vnf_instance)s '
                  '%(additional_params)s %(kwargs)s',
                  {'vnf_instance': vnf_instance,
                   'additional_params': additional_params, 'kwargs': kwargs})
        pass

    def instantiate_end(self, vnf_instance, additional_params, **kwargs):
        LOG.debug('instantiate_end %(vnf_instance)s '
                  '%(additional_params)s %(kwargs)s',
                  {'vnf_instance': vnf_instance,
                   'additional_params': additional_params, 'kwargs': kwargs})
        pass

    def terminate_start(self, vnf_instance, additional_params, **kwargs):
        LOG.debug('terminate_start %(vnf_instance)s '
                  '%(additional_params)s %(kwargs)s',
                  {'vnf_instance': vnf_instance,
                   'additional_params': additional_params, 'kwargs': kwargs})
        pass

    def terminate_end(self, vnf_instance, additional_params, **kwargs):
        LOG.debug('terminate_end %(vnf_instance)s '
                  '%(additional_params)s %(kwargs)s',
                  {'vnf_instance': vnf_instance,
                   'additional_params': additional_params, 'kwargs': kwargs})
        pass

    def scale_start(self, vnf_instance, additional_params, **kwargs):
        LOG.debug('scale_start %(vnf_instance)s '
                  '%(additional_params)s %(kwargs)s',
                  {'vnf_instance': vnf_instance,
                   'additional_params': additional_params, 'kwargs': kwargs})
        pass

    def scale_end(self, vnf_instance, additional_params, **kwargs):
        LOG.debug('scale_end %(vnf_instance)s '
                  '%(additional_params)s %(kwargs)s',
                  {'vnf_instance': vnf_instance,
                   'additional_params': additional_params, 'kwargs': kwargs})
        pass

    def heal_start(self, vnf_instance, additional_params, **kwargs):
        LOG.debug('heal_start %(vnf_instance)s '
                  '%(additional_params)s %(kwargs)s',
                  {'vnf_instance': vnf_instance,
                   'additional_params': additional_params, 'kwargs': kwargs})
        pass

    def heal_end(self, vnf_instance, additional_params, **kwargs):
        LOG.debug('heal_end %(vnf_instance)s '
                  '%(additional_params)s %(kwargs)s',
                  {'vnf_instance': vnf_instance,
                   'additional_params': additional_params, 'kwargs': kwargs})
        pass



# Copyright (c) 2019 NTT DATA.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import testtools

from tacker.common import exceptions
from tacker import context
from tacker.glance_store import store
from tacker.tests import constants


class TestStore(testtools.TestCase):

    def setUp(self):
        super(TestStore, self).setUp()
        self.context = context.get_admin_context()
        self.base_path = os.path.dirname(os.path.abspath(__file__))

    def test_get_csar_size_invalid_path(self):
        self.assertRaises(
            exceptions.VnfPackageLocationInvalid, store.get_csar_size,
            constants.UUID, 'Invalid/path')

    def test_load_csar_iter_invalid_path(self):
        self.assertRaises(
            exceptions.VnfPackageLocationInvalid, store.load_csar_iter,
            constants.UUID, 'Invalid/path')



# Copyright 2015 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""modify datatype of value

Revision ID: 5958429bcb3c
Revises: 13c0e0661015
Create Date: 2015-10-05 17:09:24.710961

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '5958429bcb3c'
down_revision = '13c0e0661015'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    op.alter_column('devicetemplateattributes',
                    'value', type_=sa.TEXT(65535), nullable=True)



# Copyright 2021 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""alter vnfd_ids, vnf_instance_ids, vnf_instance_names
columns of vnf_lcm_filters

Revision ID: 70df18f71ba2
Revises: a23ebee909a8
Create Date: 2021-06-16 14:26:51.998033

"""

# revision identifiers, used by Alembic.
revision = '70df18f71ba2'
down_revision = 'a23ebee909a8'

from alembic import op  # noqa: E402


def upgrade(active_plugins=None, options=None):
    alter_sql_vnfd_ids = "ALTER TABLE vnf_lcm_filters CHANGE \
            vnfd_ids vnfd_ids mediumtext GENERATED ALWAYS AS \
            (json_unquote(json_extract(`filter`, \
            '$.vnfInstanceSubscriptionFilter.vnfdIds'))) VIRTUAL;"

    alter_sql_vnf_instance_ids = "ALTER TABLE vnf_lcm_filters CHANGE \
            vnf_instance_ids vnf_instance_ids mediumtext GENERATED ALWAYS AS \
            (json_unquote(json_extract(`filter`, \
            '$.vnfInstanceSubscriptionFilter.vnfInstanceIds'))) VIRTUAL;"

    alter_sql_vnf_instance_names = "ALTER TABLE vnf_lcm_filters CHANGE \
                vnf_instance_names vnf_instance_names mediumtext GENERATED \
                ALWAYS AS (json_unquote(json_extract(`filter`, \
                '$.vnfInstanceSubscriptionFilter.vnfInstanceNames'))) \
                VIRTUAL;"

    op.execute(alter_sql_vnfd_ids)
    op.execute(alter_sql_vnf_instance_ids)
    op.execute(alter_sql_vnf_instance_names)



# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import testtools

from toscaparser import tosca_template
from toscaparser.utils import yamlparser
from translator.hot import tosca_translator

from tacker.tosca import utils


class TestSamples(testtools.TestCase):
    """Sample tosca validation.

    Helps to validate the tosca templates provided in the samples folder
    to make sure whether its valid YAML, valid TOSCA and
    possible to translate into HOT template.
    """

    def _get_list_of_sample(self, tosca_files):
        if tosca_files:
            base_path = (os.path.dirname(os.path.abspath(__file__)) +
                         '/../../../../samples/tosca-templates/vnfd/')
            if isinstance(tosca_files, list):
                list_of_samples = []
                for tosca_file in tosca_files:
                    sample = base_path + tosca_file
                    list_of_samples.append(sample)
                return list_of_samples

    def _test_samples(self, files):
        if files:
            for f in self._get_list_of_sample(files):
                with open(f, 'r') as _f:
                    yaml_dict = None
                    try:
                        yaml_dict = yamlparser.simple_ordered_parse(_f.read())
                    except:  # noqa
                        pass
                    self.assertIsNotNone(
                        yaml_dict,
                        "Yaml parser failed to parse %s" % f)

                    utils.updateimports(yaml_dict)

                    tosca = None
                    try:
                        tosca = tosca_template.ToscaTemplate(
                            a_file=False,
                            yaml_dict_tpl=yaml_dict)
                    except:  # noqa
                        pass

                    self.assertIsNotNone(
                        tosca,
                        "Tosca parser failed to parse %s" % f)
                    utils.post_process_template(tosca)
                    hot = None
                    try:
                        hot = tosca_translator.TOSCATranslator(tosca,
                                                               {}).translate()
                    except:  # noqa
                        pass

                    self.assertIsNotNone(
                        hot,
                        "Heat-translator failed to translate %s" % f)

    def test_scale_sample(self, tosca_file=['tosca-vnfd-scale.yaml']):
        self._test_samples(tosca_file)

    def test_alarm_sample(self, tosca_file=['tosca-vnfd-alarm-scale.yaml']):
        self._test_samples(tosca_file)

    def test_list_samples(self,
                          files=['tosca-vnfd-scale.yaml',
                                 'tosca-vnfd-alarm-scale.yaml']):
        self._test_samples(files)



# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaml

from oslo_config import cfg
from oslo_log import log as logging
from toscaparser.utils import yamlparser

from tacker.common import exceptions
from tacker.common import log
from tacker.extensions import common_services as cs
from tacker.extensions import vnfm
from tacker.vnfm.infra_drivers.kubernetes.k8s import translate_inputs
from tacker.vnfm.infra_drivers.kubernetes.k8s import translate_outputs

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class TOSCAToKubernetes(object):

    def __init__(self, vnf, core_v1_api_client,
                 app_v1_api_client, scaling_api_client):
        self.vnf = vnf
        self.core_v1_api_client = core_v1_api_client
        self.app_v1_api_client = app_v1_api_client
        self.scaling_api_client = scaling_api_client
        self.attributes = {}
        self.vnfd_yaml = None

    def generate_tosca_kube_objects(self):
        """Load TOSCA template and return tosca_kube_objects"""

        vnfd_dict = self.process_input()
        parser = translate_inputs.Parser(vnfd_dict)
        return parser.loader()

    def deploy_kubernetes_objects(self):
        """Translate tosca_kube_objects to Kubernetes objects and deploy them.

        Return a string that contains all deployment namespace and names
        """

        tosca_kube_objects = self.generate_tosca_kube_objects()
        transformer = translate_outputs.Transformer(
            core_v1_api_client=self.core_v1_api_client,
            app_v1_api_client=self.app_v1_api_client,
            scaling_api_client=self.scaling_api_client,
            k8s_client_dict=None
        )
        kubernetes_objects = transformer.transform(tosca_kube_objects)
        deployment_names = transformer.deploy(
            kubernetes_objects=kubernetes_objects)
        # return namespaces and service names for tracking resources
        return deployment_names

    def process_input(self):
        """Process input of vnfd template"""

        self.attributes = self.vnf['vnfd']['attributes'].copy()
        self.vnfd_yaml = self.attributes.pop('vnfd', None)
        if self.vnfd_yaml is None:
            LOG.error("VNFD is not provided, so no vnf is created !!")
            raise exceptions.InvalidInput("VNFD template is None.")
        LOG.debug('vnfd_yaml %s', self.vnfd_yaml)
        vnfd_dict = yamlparser.simple_ordered_parse(self.vnfd_yaml)
        LOG.debug('vnfd_dict %s', vnfd_dict)

        # Read parameter and process inputs
        if 'get_input' in str(vnfd_dict):
            self._process_parameterized_input(self.vnf['attributes'],
                                              vnfd_dict)
        return vnfd_dict

    @log.log
    def _update_params(self, original, paramvalues):
        for key, value in (original).items():
            if not isinstance(value, dict) or 'get_input' not in str(value):
                pass
            elif isinstance(value, dict):
                if 'get_input' in value:
                    if value['get_input'] in paramvalues:
                        original[key] = paramvalues[value['get_input']]
                    else:
                        LOG.debug('Key missing Value: %s', key)
                        raise cs.InputValuesMissing(key=key)
                else:
                    self._update_params(value, paramvalues)

    @log.log
    def _process_parameterized_input(self, attrs, vnfd_dict):
        param_vattrs_yaml = attrs.pop('param_values', None)
        if param_vattrs_yaml:
            try:
                param_vattrs_dict = yaml.safe_load(param_vattrs_yaml)
                LOG.debug('param_vattrs_yaml', param_vattrs_dict)
                for node in vnfd_dict['topology_template']['node_templates'].\
                        values():
                    if 'get_input' in str(node):
                        self._update_params(node, param_vattrs_dict)
            except Exception as e:
                LOG.debug("Not Well Formed: %s", str(e))
                raise vnfm.ParamYAMLNotWellFormed(
                    error_msg_details=str(e))
            else:
                self._update_params(vnfd_dict, param_vattrs_dict)
        else:
            raise cs.ParamYAMLInputMissing()



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from tacker.sol_refactored.api.policies import vnflcm_v2 as vnflcm_policy_v2
from tacker.sol_refactored.api import wsgi as sol_wsgi
from tacker.sol_refactored.controller import vnflcm_v2
from tacker.sol_refactored.controller import vnflcm_versions


class VnflcmVersions(sol_wsgi.SolAPIRouter):

    controller = sol_wsgi.SolResource(
        vnflcm_versions.VnfLcmVersionsController())
    route_list = [("/api_versions", {"GET": "index"})]


class VnflcmAPIRouterV2(sol_wsgi.SolAPIRouter):

    controller = sol_wsgi.SolResource(vnflcm_v2.VnfLcmControllerV2(),
                                      policy_name=vnflcm_policy_v2.POLICY_NAME)
    route_list = [
        ("/vnf_instances", {"GET": "index", "POST": "create"}),
        ("/vnf_instances/{id}",
         {"DELETE": "delete", "GET": "show", "PATCH": "update"}),
        ("/vnf_instances/{id}/instantiate", {"POST": "instantiate"}),
        ("/vnf_instances/{id}/heal", {"POST": "heal"}),
        ("/vnf_instances/{id}/terminate", {"POST": "terminate"}),
        ("/vnf_instances/{id}/scale", {"POST": "scale"}),
        ("/vnf_instances/{id}/change_ext_conn", {"POST": "change_ext_conn"}),
        ("/api_versions", {"GET": "api_versions"}),
        ("/subscriptions", {"GET": "subscription_list",
                            "POST": "subscription_create"}),
        ("/subscriptions/{id}", {"GET": "subscription_show",
                                 "DELETE": "subscription_delete"}),
        ("/vnf_lcm_op_occs", {"GET": "lcm_op_occ_list"}),
        ("/vnf_lcm_op_occs/{id}/retry", {"POST": "lcm_op_occ_retry"}),
        ("/vnf_lcm_op_occs/{id}/rollback", {"POST": "lcm_op_occ_rollback"}),
        ("/vnf_lcm_op_occs/{id}/fail", {"POST": "lcm_op_occ_fail"}),
        # NOTE: 'DELETE' is not defined in the specification. It is for test
        # use since it is convenient to be able to delete under development.
        # It is available when config parameter
        # v2_vnfm.test_enable_lcm_op_occ_delete set to True.
        ("/vnf_lcm_op_occs/{id}", {"GET": "lcm_op_occ_show",
                                   "DELETE": "lcm_op_occ_delete"})
    ]



# Copyright 2011 VMware, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging
from oslo_service import periodic_task

from tacker._i18n import _
from tacker.common import utils


LOG = logging.getLogger(__name__)


class Manager(periodic_task.PeriodicTasks):

    # Set RPC API version to 1.0 by default.
    target = oslo_messaging.Target(version='1.0')

    def __init__(self, host=None):
        if not host:
            host = cfg.CONF.host
        self.host = host
        conf = getattr(self, "conf", cfg.CONF)
        super(Manager, self).__init__(conf)

    def periodic_tasks(self, context, raise_on_error=False):
        self.run_periodic_tasks(context, raise_on_error=raise_on_error)

    def start(self):
        """Start service."""
        pass

    def stop(self):
        """Stop service."""
        pass

    def init_host(self):
        """Handle initialization if this is a standalone service.

        Child classes should override this method.

        """
        pass

    def after_start(self):
        """Handler post initialization stuff.

        Child classes can override this method.
        """
        pass


def validate_post_plugin_load():
    """Checks if the configuration variables are valid.

    If the configuration is invalid then the method will return an error
    message. If all is OK then it will return None.
    """
    message = None
    return message


def validate_pre_plugin_load():
    """Checks if the configuration variables are valid.

    If the configuration is invalid then the method will return an error
    message. If all is OK then it will return None.
    """
    message = None
    return message


class TackerManager(object):
    """Tacker's Manager class.

    Tacker's Manager class is responsible for parsing a config file and
    instantiating the correct plugin that concretely implement
    tacker_plugin_base class.
    The caller should make sure that TackerManager is a singleton.
    """
    _instance = None

    def __init__(self, options=None, config_file=None):
        # If no options have been provided, create an empty dict
        if not options:
            options = {}

        msg = validate_pre_plugin_load()
        if msg:
            LOG.critical(msg)
            raise Exception(msg)

        msg = validate_post_plugin_load()
        if msg:
            LOG.critical(msg)
            raise Exception(msg)

        self.service_plugins = {}
        self._load_service_plugins()

    @staticmethod
    def load_class_for_provider(namespace, plugin_provider):
        """Loads plugin using alias or class name

        Load class using stevedore alias or the class name
        :param namespace: namespace where alias is defined
        :param plugin_provider: plugin alias or class name
        :returns: plugin that is loaded
        :raises ImportError: if fails to load plugin
        """

        try:
            return utils.load_class_by_alias_or_classname(namespace,
                    plugin_provider)
        except ImportError:
            raise ImportError(_("Plugin '%s' not found.") % plugin_provider)

    def _get_plugin_instance(self, namespace, plugin_provider):
        plugin_class = self.load_class_for_provider(namespace, plugin_provider)
        return plugin_class()

    def _load_service_plugins(self):
        """Loads service plugins.

        Starts from the core plugin and checks if it supports
        advanced services then loads classes provided in configuration.
        """
        plugin_providers = cfg.CONF.service_plugins
        if 'commonservices' not in plugin_providers:
            plugin_providers.append('commonservices')
        LOG.debug("Loading service plugins: %s", plugin_providers)
        for provider in plugin_providers:
            if provider == '':
                continue
            LOG.info("Loading Plugin: %s", provider)

            plugin_inst = self._get_plugin_instance('tacker.service_plugins',
                                                    provider)
            # only one implementation of svc_type allowed
            # specifying more than one plugin
            # for the same type is a fatal exception
            if plugin_inst.get_plugin_type() in self.service_plugins:
                raise ValueError(_("Multiple plugins for service "
                                   "%s were configured"),
                                 plugin_inst.get_plugin_type())

            self.service_plugins[plugin_inst.get_plugin_type()] = plugin_inst
            # # search for possible agent notifiers declared in service plugin
            # # (needed by agent management extension)
            # if (hasattr(self.plugin, 'agent_notifiers') and
            #         hasattr(plugin_inst, 'agent_notifiers')):
            #     self.plugin.agent_notifiers.update(plugin_inst.agent_notifiers)

            LOG.debug("Successfully loaded %(type)s plugin. "
                      "Description: %(desc)s",
                      {"type": plugin_inst.get_plugin_type(),
                       "desc": plugin_inst.get_plugin_description()})

    @classmethod
    @utils.synchronized("manager")
    def _create_instance(cls):
        if cls._instance is None:
            cls._instance = cls()

    @classmethod
    def get_instance(cls):
        # double checked locking
        if cls._instance is None:
            cls._create_instance()
        return cls._instance

    @classmethod
    def get_plugin(cls):
        return cls.get_instance().plugin

    @classmethod
    def get_service_plugins(cls):
        return cls.get_instance().service_plugins

    @classmethod
    def has_instance(cls):
        return cls._instance is not None

    @classmethod
    def clear_instance(cls):
        cls._instance = None



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from mistralclient.api import client as mistral_client


class MistralClient(object):
    """Mistral Client class for NSD"""

    def __init__(self, keystone, auth_token):
        endpoint = keystone.session.get_endpoint(
            service_type='workflowv2', region_name=None)

        self.client = mistral_client.client(auth_token=auth_token,
            mistral_url=endpoint)

    def get_client(self):
        return self.client



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tacker.sol_refactored.objects import base
from tacker.sol_refactored.objects import fields


# NFV-SOL 003
# - v3.3.1 9.5.3.12 (API version: 1.4.0)
@base.TackerObjectRegistry.register
class VimSnapshotResourceV1(base.TackerObject, base.TackerObjectDictCompat):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'vimConnectionId': fields.StringField(nullable=True),
        'resourceProviderId': fields.StringField(nullable=True),
        'vnfSnapshotId': fields.StringField(nullable=False),
        'vnfcSnapshotId': fields.StringField(nullable=False),
        'storageSnapshotId': fields.StringField(nullable=True),
        'vimSnapshotResourceId': fields.StringField(nullable=False),
    }



# Copyright 2019 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""add size column to vnf_packages table

Revision ID: d2e39e01d540
Revises: 985e28392890
Create Date: 2019-11-27 13:30:23.599865

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'd2e39e01d540'
down_revision = '985e28392890'


# Added 'size' column to an existing table. Server_default bit will make
# existing rows 0 for that column
def upgrade(active_plugins=None, options=None):
    op.add_column('vnf_packages', sa.Column('size', sa.BigInteger,
                                           nullable=False,
                                           server_default='0'))



# Copyright (c) 2013 Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr

from testtools import matchers
from webob import exc

from oslo_policy import policy as oslo_policy
from oslo_serialization import jsonutils
from tacker.api import api_common as common
from tacker.api.v1 import resource as wsgi_resource
from tacker.common import exceptions
from tacker.tests import base


class FakeController(common.TackerController):
    _resource_name = 'fake'


class APICommonTestCase(base.BaseTestCase):
    def setUp(self):
        super(APICommonTestCase, self).setUp()
        self.controller = FakeController(None)

    def test_prepare_request_body(self):
        body = {
            'fake': {
                'name': 'terminator',
                'model': 'T-800',
            }
        }
        params = [
            {'param-name': 'name',
             'required': True},
            {'param-name': 'model',
             'required': True},
            {'param-name': 'quote',
             'required': False,
             'default-value': "i'll be back"},
        ]
        expect = {
            'fake': {
                'name': 'terminator',
                'model': 'T-800',
                'quote': "i'll be back",
            }
        }
        actual = self.controller._prepare_request_body(body, params)
        self.assertThat(expect, matchers.Equals(actual))

    def test_prepare_request_body_none(self):
        body = None
        params = [
            {'param-name': 'quote',
             'required': False,
             'default-value': "I'll be back"},
        ]
        expect = {
            'fake': {
                'quote': "I'll be back",
            }
        }
        actual = self.controller._prepare_request_body(body, params)
        self.assertThat(expect, matchers.Equals(actual))

    def test_prepare_request_body_keyerror(self):
        body = {'t2': {}}
        params = []
        self.assertRaises(exc.HTTPBadRequest,
                          self.controller._prepare_request_body,
                          body,
                          params)

    def test_prepare_request_param_value_none(self):
        body = {
            'fake': {
                'name': None,
            }
        }
        params = [
            {'param-name': 'name',
             'required': True},
        ]
        self.assertRaises(exc.HTTPBadRequest,
                          self.controller._prepare_request_body,
                          body,
                          params)

    def test_http_client_exception(self):
        req = wsgi_resource.Request({})
        language = req.best_match_language()
        e = exc.HTTPClientError()
        result = common.convert_exception_to_http_exc(e, {}, language)

        except_res = {'message': 'The server could not comply with'
                                 ' the request since it is either '
                                 'malformed or otherwise incorrect.',
                      'type': 'HTTPClientError',
                      'detail': ''}

        self.assertEqual(
            except_res, jsonutils.loads(result.body)["TackerError"])
        self.assertEqual(400, result.code)

    def test_http_exception(self):
        req = wsgi_resource.Request({})
        language = req.best_match_language()
        e = exc.HTTPException
        result = common.convert_exception_to_http_exc(e, {}, language)

        except_res = {"message": "Request Failed: internal server error "
                                 "while processing your request.",
                      "type": "HTTPInternalServerError",
                      "detail": ""}

        self.assertEqual(
            except_res, jsonutils.loads(result.body)["TackerError"])
        self.assertEqual(500, result.code)

    def test_tacker_exception(self):
        req = wsgi_resource.Request({})
        language = req.best_match_language()
        e = exceptions.TackerException()
        result = common.convert_exception_to_http_exc(e, {}, language)

        except_res = {'message': 'An unknown exception occurred.',
                      'type': 'TackerException',
                      'detail': ''}

        self.assertEqual(
            except_res, jsonutils.loads(result.body)["TackerError"])
        self.assertEqual(500, result.code)

    def test_addr_format_error_exception(self):
        req = wsgi_resource.Request({})
        language = req.best_match_language()
        e = netaddr.AddrFormatError()
        result = common.convert_exception_to_http_exc(e, {}, language)

        except_res = {'message': '',
                      'type': 'AddrFormatError',
                      'detail': ''}

        self.assertEqual(
            except_res, jsonutils.loads(result.body)["TackerError"])
        self.assertEqual(500, result.code)

    def test_policy_not_authorized_exception(self):
        req = wsgi_resource.Request({})
        language = req.best_match_language()
        e = oslo_policy.PolicyNotAuthorized(None, None, None)
        result = common.convert_exception_to_http_exc(e, {}, language)

        except_res = {'message': 'None is disallowed by policy',
                      'type': 'PolicyNotAuthorized',
                      'detail': ''}

        self.assertEqual(
            except_res, jsonutils.loads(result.body)["TackerError"])
        self.assertEqual(500, result.code)

    def test_not_implemented_error_exception(self):
        req = wsgi_resource.Request({})
        language = req.best_match_language()
        e = NotImplementedError()
        result = common.convert_exception_to_http_exc(e, {}, language)

        except_res = {'NotImplementedError': {'message': '',
                                              'type': 'NotImplementedError',
                                              'detail': ''}}

        self.assertEqual(except_res, jsonutils.loads(result.body))
        self.assertEqual(501, result.code)

    def test_get_exception_data(self):
        result = common.get_exception_data(exc.HTTPClientError)
        self.assertEqual("", result["detail"])
        self.assertEqual("type", result["type"])
        self.assertEqual(exc.HTTPClientError, result["message"])


class SortingEmulatedHelperTestcase(base.BaseTestCase):

    class request(object):
        class GET(object):
            @staticmethod
            def getall(cont):
                if cont == "sort_key":
                    return {"sort_key": "test"}
                return {'asc': "test"}

    def test_sort_emulate_class(self):
        sort = common.SortingEmulatedHelper(self.request, ["sort_key"])
        self.assertEqual([('sort_key', True)], list(sort.sort_dict))

    def test_update_fields(self):
        sort = common.SortingEmulatedHelper(self.request, ["sort_key"])
        original_fields = ["sort_dir"]
        fields_to_add = []
        sort.update_fields(original_fields, fields_to_add)
        self.assertEqual(["sort_dir", "sort_key"], original_fields)
        self.assertEqual(["sort_key"], fields_to_add)

    def test_sort(self):
        sort = common.SortingEmulatedHelper(self.request, ["sort_key"])
        items = [{"sort_key": 9}, {"sort_key": 6}, {"sort_key": 7}]
        expect_res = [{"sort_key": 6}, {"sort_key": 7}, {"sort_key": 9}]
        result = sort.sort(items)
        self.assertEqual(expect_res, result)

        items = [{"sort_key": 9, "sort_key_1": None},
                 {"sort_key": 6, "sort_key_2": {"sort_key_3": 8,
                                                "sort_key": 3}},
                 {"sort_key": 7}]
        expect_res = [{"sort_key": 6, "sort_key_2": {"sort_key": 3,
                                                     "sort_key_3": 8}},
                      {"sort_key": 7},
                      {"sort_key": 9, "sort_key_1": None}]
        result = sort.sort(items)
        self.assertEqual(expect_res, result)


class PaginationEmulatedHelperTestcase(base.BaseTestCase):

    class request(object):
        path_url = "http://127.0.0.1"

        class GET(object):

            @staticmethod
            def get(limit, reverse):
                if limit == "page_reverse":
                    return "True"
                return 5

            @staticmethod
            def copy():
                return {}

    class request_revs(object):
        path_url = "http://127.0.0.1"

        class GET(object):

            @staticmethod
            def get(limit, reverse):
                if limit == "page_reverse":
                    return "False"
                return 5

            @staticmethod
            def copy():
                return {}

    def test_pagination_emulate_class(self):
        page = common.PaginationEmulatedHelper(self.request)
        self.assertEqual(5, page.limit)
        self.assertEqual(True, page.page_reverse)
        self.assertEqual(5, page.marker)

    def test_update_fields(self):
        page = common.PaginationEmulatedHelper(self.request)
        original_fields = ["name"]
        fields_to_add = []
        page.update_fields(original_fields, fields_to_add)
        self.assertEqual(["name", "id"], original_fields)
        self.assertEqual(["id"], fields_to_add)

    def test_paginate(self):
        page = common.PaginationEmulatedHelper(self.request)
        items = [{"id": 6}, {"id": 4}, {"id": 2}]
        expect_res = [{"id": 6}, {"id": 4}]
        result = page.paginate(items)
        self.assertEqual(expect_res, result)

        items = [{"id": 6}, {"id": 4}, {"id": 2}, {"id": 7},
                 {"id": 3}, {"id": 8}, {"id": 5}]
        expect_res = [{"id": 4}, {"id": 2}, {"id": 7}, {"id": 3}, {"id": 8}]
        result = page.paginate(items)
        self.assertEqual(expect_res, result)

        items = [{"id": 6}, {"id": 4}, {"id": 5}, {"id": 7},
                 {"id": 3}, {"id": 8}, {"id": 2}]
        result = page.paginate(items)
        self.assertEqual([], result)

    def test_get_links(self):
        page = common.PaginationEmulatedHelper(self.request)
        items = [{"id": 6}, {"id": 4}, {"id": 2}]
        expect_res = [{'href': 'http://127.0.0.1?marker=2', 'rel': 'next'}]
        result = page.get_links(items)
        self.assertEqual(expect_res, result)

        page_previous = common.PaginationEmulatedHelper(self.request_revs)
        items = [{"id": 6}, {"id": 4}, {"id": 2}]
        expect_href = ('http://127.0.0.1?page_reverse=True&marker=6',
                       'http://127.0.0.1?marker=6&page_reverse=True')
        result = page_previous.get_links(items)

        self.assertEqual('previous', result[0]['rel'])
        self.assertIn(result[0]['href'], expect_href)



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""uuid consistency

Revision ID: f958f58e5daa
Revises: acf941e54075
Create Date: 2016-05-28 07:13:07.125562

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = 'f958f58e5daa'
down_revision = 'acf941e54075'


from alembic import op

from tacker.db import migration
from tacker.db import types

FK_MAP = {'vims': ('vimauths', 'devices'), 'devices': ('deviceattributes',
          'proxymgmtports'), 'devicetemplates': ('devices', 'servicetypes',
                                                 'devicetemplateattributes')}


def upgrade(active_plugins=None, options=None):

    pk_id_tables = ('vims', 'vimauths', 'devices', 'deviceattributes',
                  'servicetypes', 'devicetemplates',
                  'devicetemplateattributes')
    for table in pk_id_tables:
        with migration.modify_foreign_keys_constraint(FK_MAP.get(table, [])):
            op.alter_column(table, 'id', type_=types.Uuid,
                            nullable=False)

    fk_template_id_tables = ('devices', 'servicetypes',
                             'devicetemplateattributes')
    for table in fk_template_id_tables:
        with migration.modify_foreign_keys_constraint(fk_template_id_tables):
            op.alter_column(table, 'template_id', type_=types.Uuid,
                            nullable=False)

    fk_vim_id_tables = ('devices', 'vimauths')
    for table in fk_vim_id_tables:
        with migration.modify_foreign_keys_constraint(fk_vim_id_tables):
            op.alter_column(table, 'vim_id', type_=types.Uuid,
                            nullable=False)

    fk_device_id_tables = ('deviceattributes', 'proxymgmtports')
    for table in fk_device_id_tables:
        with migration.modify_foreign_keys_constraint(fk_device_id_tables):
            op.alter_column(table, 'device_id', type_=types.Uuid,
                            nullable=False)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tacker.extensions import vnfm
from tacker.vnfm.lcm_user_data.abstract_user_data import AbstractUserData


class SampleUserData(AbstractUserData):

    @staticmethod
    def instantiate(base_hot_dict=None,
                    vnfd_dict=None,
                    inst_req_info=None,
                    grant_info=None):

        error_reason = _(
            "invalid user data script.")
        raise vnfm.LCMUserDataFailed(reason=error_reason)



#    Copyright 2018 NTT DATA.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import datetime

import oslo_messaging as messaging
from oslo_utils import versionutils
from oslo_versionedobjects import base as ovoo_base
from oslo_versionedobjects import exception as ovoo_exc

from tacker import objects
from tacker.objects import fields as obj_fields


def get_attrname(name):
    """Return the mangled name of the attribute's underlying storage."""
    return '_obj_' + name


class TackerObjectRegistry(ovoo_base.VersionedObjectRegistry):
    notification_classes = []

    def registration_hook(self, cls, index):
        # NOTE(bhagyashris): This is called when an object is registered,
        # and is responsible for maintaining tacker.objects.$OBJECT
        # as the highest-versioned implementation of a given object.
        version = versionutils.convert_version_to_tuple(cls.VERSION)
        if not hasattr(objects, cls.obj_name()):
            setattr(objects, cls.obj_name(), cls)
        else:
            cur_version = versionutils.convert_version_to_tuple(
                getattr(objects, cls.obj_name()).VERSION)
            if version >= cur_version:
                setattr(objects, cls.obj_name(), cls)


class TackerObject(ovoo_base.VersionedObject):
    # NOTE(bhagyashris): OBJ_PROJECT_NAMESPACE needs to be set so that nova,
    # tacker, and other objects can exist on the same bus and be distinguished
    # from one another.
    OBJ_SERIAL_NAMESPACE = 'tacker_object'
    OBJ_PROJECT_NAMESPACE = 'tacker'

    def __init__(self, context=None, **kwargs):
        super(TackerObject, self).__init__(context, **kwargs)
        self.obj_set_defaults()

    def tacker_obj_get_changes(self):
        """Returns a dict of changed fields with tz unaware datetimes.

        Any timezone aware datetime field will be converted to UTC timezone
        and returned as timezone unaware datetime.

        This will allow us to pass these fields directly to a db update
        method as they can't have timezone information.
        """
        # Get dirtied/changed fields
        changes = self.obj_get_changes()

        # Look for datetime objects that contain timezone information
        for k, v in changes.items():
            if isinstance(v, datetime.datetime) and v.tzinfo:
                # Remove timezone information and adjust the time according to
                # the timezone information's offset.
                changes[k] = v.replace(tzinfo=None) - v.utcoffset()

        # Return modified dict
        return changes

    def obj_reset_changes(self, fields=None, recursive=False):
        """Reset the list of fields that have been changed.

        .. note::

          - This is NOT "revert to previous values"
          - Specifying fields on recursive resets will only be honored at the
            top level. Everything below the top will reset all.

        :param fields: List of fields to reset, or "all" if None.
        :param recursive: Call obj_reset_changes(recursive=True) on
                          any sub-objects within the list of fields
                          being reset.
        """
        if recursive:
            for field in self.obj_get_changes():

                # Ignore fields not in requested set (if applicable)
                if fields and field not in fields:
                    continue

                # Skip any fields that are unset
                if not self.obj_attr_is_set(field):
                    continue

                value = getattr(self, field)

                # Don't reset nulled fields
                if value is None:
                    continue

                # Reset straight Object and ListOfObjects fields
                if isinstance(self.fields[field], obj_fields.ObjectField):
                    value.obj_reset_changes(recursive=True)
                elif isinstance(self.fields[field],
                                obj_fields.ListOfObjectsField):
                    for thing in value:
                        thing.obj_reset_changes(recursive=True)

        if fields:
            self._changed_fields -= set(fields)
        else:
            self._changed_fields.clear()

    @contextlib.contextmanager
    def obj_alternate_context(self, context):
        original_context = self._context
        self._context = context
        try:
            yield
        finally:
            self._context = original_context


class TackerObjectSerializer(messaging.NoOpSerializer):
    """A TackerObject-aware Serializer.

    This implements the Oslo Serializer interface and provides the
    ability to serialize and deserialize TackerObject entities. Any service
    that needs to accept or return TackerObjects as arguments or result values
    should pass this to its RPCClient and RPCServer objects.
    """

    def _process_object(self, context, objprim):
        try:
            objinst = TackerObject.obj_from_primitive(objprim, context=context)
        except ovoo_exc.IncompatibleObjectVersion:
            raise
        return objinst

    def _process_iterable(self, context, action_fn, values):
        """Process an iterable, taking an action on each value.

        :param:context: Request context
        :param:action_fn: Action to take on each item in values
        :param:values: Iterable container of things to take action on
        :returns: A new container of the same type (except set) with
                  items from values having had action applied.
        """

        iterable = values.__class__
        if issubclass(iterable, dict):
            return iterable(**{k: action_fn(context, v)
                               for k, v in values.items()})
        else:
            # NOTE(nirajsingh) A set can't have an unhashable value inside,
            # such as a dict. Convert the set to list, which is fine, since we
            # can't send them over RPC anyway. We convert it to list as this
            # way there will be no semantic change between the fake rpc driver
            # used in functional test and a normal rpc driver.
            if iterable == set:
                iterable = list
            return iterable([action_fn(context, value) for value in values])

    def serialize_entity(self, context, entity):
        if isinstance(entity, (tuple, list, set, dict)):
            entity = self._process_iterable(context, self.serialize_entity,
                                            entity)
        elif (hasattr(entity, 'obj_to_primitive') and
              callable(entity.obj_to_primitive)):
            entity = entity.obj_to_primitive()
        return entity

    def deserialize_entity(self, context, entity):
        if isinstance(entity, dict) and 'tacker_object.name' in entity:
            entity = self._process_object(context, entity)
        elif isinstance(entity, (tuple, list, set, dict)):
            entity = self._process_iterable(context, self.deserialize_entity,
                                            entity)
        return entity


class TackerPersistentObject(object):
    """Mixin class for Persistent objects.

    This adds the fields that we use in common for most persistent objects.
    """
    fields = {
        'created_at': obj_fields.DateTimeField(nullable=False),
        'updated_at': obj_fields.DateTimeField(nullable=True),
        'deleted_at': obj_fields.DateTimeField(nullable=True),
        'deleted': obj_fields.BooleanField(default=False)
    }


remotable = ovoo_base.remotable
remotable_classmethod = ovoo_base.remotable_classmethod
obj_make_list = ovoo_base.obj_make_list
TackerObjectDictCompat = ovoo_base.VersionedObjectDictCompat



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""make VNFD/VNF/VIM name mandatory

Revision ID: 5f88e86b35c7
Revises: 354de64ba129
Create Date: 2016-06-14 11:16:16.303343

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '5f88e86b35c7'
down_revision = '354de64ba129'

from alembic import op
from sqlalchemy.dialects import mysql


def upgrade(active_plugins=None, options=None):
    op.alter_column('devices', 'name',
               existing_type=mysql.VARCHAR(length=255),
               nullable=False)
    op.alter_column('devicetemplates', 'name',
               existing_type=mysql.VARCHAR(length=255),
               nullable=False)
    op.alter_column('vims', 'name',
               existing_type=mysql.VARCHAR(length=255),
               nullable=False)



#    Copyright 2020 NTT DATA.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_log import log as logging
from oslo_versionedobjects import base as ovoo_base
from tacker._i18n import _
from tacker.common import exceptions
from tacker.common import utils
from tacker.db import api as db_api
from tacker.db.db_sqlalchemy import api
from tacker.db.db_sqlalchemy import models
from tacker.objects import base
from tacker.objects import fields

LOG = logging.getLogger(__name__)


@db_api.context_manager.writer
def _vnf_artifacts_create(context, values):
    vnf_artifacts = models.VnfPackageArtifactInfo()

    vnf_artifacts.update(values)
    vnf_artifacts.save(context.session)

    return vnf_artifacts


@db_api.context_manager.reader
def _vnf_artifact_get_by_id(context, id):

    query = api.model_query(context, models.VnfPackageArtifactInfo,
                            read_deleted="no").filter_by(id=id)

    result = query.first()

    if not result:
        raise exceptions.VnfArtifactNotFound(id=id)

    return result


@base.TackerObjectRegistry.register
class VnfPackageArtifactInfo(base.TackerObject, base.TackerPersistentObject):
    ALL_ATTRIBUTES = {
        "additionalArtifacts": {
            'artifactPath': ('artifact_path', 'string',
                             'VnfPackageArtifactInfo'),
            'metadata': ('_metadata', 'dict', 'VnfPackageArtifactInfo'),
            "checksum": {
                'hash': ('hash', 'string', 'VnfPackageArtifactInfo'),
                'algorithm': ('algorithm', 'string', 'VnfPackageArtifactInfo')
            }
        }
    }

    FLATTEN_ATTRIBUTES = utils.flatten_dict(ALL_ATTRIBUTES.copy())
    SIMPLE_ATTRIBUTES = ['artifactPath']
    COMPLEX_ATTRIBUTES = [
        'additionalArtifacts',
        'additionalArtifacts/metadata',
        'additionalArtifacts/checksum']

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(nullable=False),
        'package_uuid': fields.UUIDField(nullable=False),
        'artifact_path': fields.StringField(nullable=False),
        'algorithm': fields.StringField(nullable=False),
        'hash': fields.StringField(nullable=False),
        '_metadata': fields.DictOfStringsField(nullable=True, default={})
    }

    @base.remotable_classmethod
    def get_by_id(cls, context, id):
        db_artifact = _vnf_artifact_get_by_id(context, id)
        return cls._from_db_object(context, cls(), db_artifact)

    @staticmethod
    def _from_db_object(context, vnf_artifacts, db_vnf_artifacts):

        for key in vnf_artifacts.fields:
            setattr(vnf_artifacts, key, db_vnf_artifacts[key])

        vnf_artifacts._context = context
        vnf_artifacts.obj_reset_changes()

        return vnf_artifacts

    def obj_load_attr(self, attrname):
        if not self._context:
            raise exceptions.OrphanedObjectError(
                method='obj_load_attr', objtype=self.obj_name())
        if 'id' not in self:
            raise exceptions.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)

        LOG.debug("Lazy-loading '%(attr)s' on %(name)s id %(id)s",
                  {'attr': attrname,
                   'name': self.obj_name(),
                   'id': self.id,
                   })

        self._obj_load_attr(attrname)

    def _obj_load_attr(self, attrname):

        if attrname in self.fields and attrname != 'id':
            self._load_generic(attrname)
        else:
            # NOTE(nirajsingh): Raise error if non existing field is
            # requested.
            raise exceptions.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)

        self.obj_reset_changes([attrname])

    def _load_generic(self, attrname):
        artifact = self.__class__.get_by_id(self._context,
                                            id=self.id)
        if attrname not in artifact:
            raise exceptions.ObjectActionError(
                action='obj_load_attr',
                reason=_('loading %s requires recursion') % attrname)

        for field in self.fields:
            if field in artifact and field not in self:
                setattr(self, field, getattr(artifact, field))

    @base.remotable
    def create(self):
        if self.obj_attr_is_set('id'):
            raise exceptions.ObjectActionError(action='create',
                                               reason=_('already created'))

        updates = self.obj_get_changes()
        db_vnf_artifacts = _vnf_artifacts_create(
            self._context, updates)
        self._from_db_object(self._context, self, db_vnf_artifacts)

    def to_dict(self, include_fields=None):
        response = dict()
        fields = ['additionalArtifacts/%s' % attribute for attribute in
            self.SIMPLE_ATTRIBUTES]

        to_fields = set(fields).intersection(include_fields)
        for field in to_fields:
            display_field = field.split("/")[-1]
            response[display_field] = getattr(
                self, self.FLATTEN_ATTRIBUTES[field][0])

        to_fields = set([key for key in self.FLATTEN_ATTRIBUTES.keys()
            if key.startswith('additionalArtifacts/checksum')])
        checksum = dict()
        to_fields = to_fields.intersection(include_fields)
        for field in to_fields:
            display_field = field.split("/")[-1]
            checksum[display_field] = getattr(
                self, self.FLATTEN_ATTRIBUTES[field][0])

        if checksum:
            response.update({"checksum": checksum})

        metadata = dict()
        to_fields = set(['additionalArtifacts/metadata']).\
            intersection(include_fields)
        if to_fields:
            metadata_json = \
                getattr(self, self.
                        FLATTEN_ATTRIBUTES['additionalArtifacts/metadata'][0])
            if metadata_json is not None:
                metadata.update(metadata_json)

            response.update({"metadata": metadata})

        return response


@base.TackerObjectRegistry.register
class VnfPackageArtifactInfoList(ovoo_base.ObjectListBase, base.TackerObject):
    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('VnfPackageArtifactInfo')
    }

    def to_dict(self, include_fields=None):
        artifactList = list()
        for artifact in self.objects:
            arti_dict = artifact.to_dict(include_fields)

            if arti_dict:
                artifactList.append(arti_dict)

        return artifactList



#    Copyright 2019 NTT DATA.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import timeutils
from oslo_utils import uuidutils
from oslo_versionedobjects import base as ovoo_base
from sqlalchemy.orm import joinedload

from tacker.common import exceptions
from tacker.db import api as db_api
from tacker.db.db_sqlalchemy import api
from tacker.db.db_sqlalchemy import models
from tacker import objects
from tacker.objects import base
from tacker.objects import fields


_NO_DATA_SENTINEL = object()

VNF_DEPLOYMENT_FLAVOUR_OPTIONAL_ATTRS = ['software_images']

LOG = logging.getLogger(__name__)


@db_api.context_manager.writer
def _vnf_deployment_flavour_create(context, values):
    vnf_deployment_flavour = models.VnfDeploymentFlavour()

    vnf_deployment_flavour.update(values)
    vnf_deployment_flavour.save(context.session)

    return vnf_deployment_flavour


@db_api.context_manager.reader
def _vnf_deployment_flavour_get_by_id(context, id, columns_to_join=None):

    query = api.model_query(context, models.VnfDeploymentFlavour,
                            read_deleted="no").filter_by(id=id)

    if columns_to_join:
        for column in columns_to_join:
            query = query.options(joinedload(column))

    result = query.first()

    if not result:
        raise exceptions.VnfDeploymentFlavourNotFound(id=id)

    return result


@db_api.context_manager.writer
def _destroy_vnf_deployment_flavour(context, flavour_uuid):
    now = timeutils.utcnow()
    updated_values = {'deleted': True,
                      'deleted_at': now
                      }

    software_images_query = api.model_query(
        context, models.VnfSoftwareImage,
        (models.VnfSoftwareImage.id,)).filter_by(flavour_uuid=flavour_uuid)

    api.model_query(context, models.VnfSoftwareImageMetadata). \
        filter(models.VnfSoftwareImageMetadata.image_uuid.
        in_(software_images_query.subquery())).update(
        updated_values, synchronize_session=False)

    api.model_query(context, models.VnfSoftwareImage). \
        filter_by(flavour_uuid=flavour_uuid). \
        update(updated_values, synchronize_session=False)
    api.model_query(context, models.VnfDeploymentFlavour). \
        filter_by(id=flavour_uuid). \
        update(updated_values, synchronize_session=False)


@base.TackerObjectRegistry.register
class VnfDeploymentFlavour(base.TackerObject, base.TackerPersistentObject):

    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(nullable=False),
        'package_uuid': fields.UUIDField(nullable=False),
        'flavour_id': fields.StringField(nullable=False),
        'flavour_description': fields.StringField(nullable=False),
        'instantiation_levels': fields.DictOfNullableField(nullable=True),
        'software_images': fields.ObjectField('VnfSoftwareImagesList'),
    }

    @staticmethod
    def _from_db_object(context, flavour, db_flavour, expected_attrs=None):
        flavour._context = context

        special_cases = set(['instantiation_levels'])
        fields = set(flavour.fields) - special_cases

        for key in fields:
            if key in VNF_DEPLOYMENT_FLAVOUR_OPTIONAL_ATTRS:
                continue
            if db_flavour[key]:
                setattr(flavour, key, db_flavour[key])

        inst_levels = db_flavour['instantiation_levels']
        if inst_levels:
            flavour.instantiation_levels = jsonutils.loads(inst_levels)

        flavour._extra_attributes_from_db_object(flavour, db_flavour,
                                                expected_attrs)

        flavour.obj_reset_changes()

        return flavour

    @staticmethod
    def _extra_attributes_from_db_object(flavour, db_flavour,
                                         expected_attrs=None):
        """Method to help with migration of extra attributes to objects.

        """
        if expected_attrs is None:
            expected_attrs = []

        if 'software_images' in expected_attrs:
            flavour._load_sw_images(db_flavour.get('software_images'))

    @base.remotable
    def create(self):
        if self.obj_attr_is_set('id'):
            raise exceptions.ObjectActionError(action='create',
                                               reason='already created')
        updates = self.obj_get_changes()

        if 'id' not in updates:
            updates['id'] = uuidutils.generate_uuid()
            self.id = updates['id']

        if 'software_images' in updates:
            updates.pop('software_images')

        special_key = 'instantiation_levels'
        if special_key in updates:
            updates[special_key] = jsonutils.dumps(updates.get(special_key))

        db_flavour = _vnf_deployment_flavour_create(self._context, updates)
        self._from_db_object(self._context, self, db_flavour)

    @base.remotable_classmethod
    def get_by_id(cls, context, id, expected_attrs=None):
        db_flavour = _vnf_deployment_flavour_get_by_id(
            context, id, columns_to_join=expected_attrs)
        return cls._from_db_object(context, cls(), db_flavour,
                                   expected_attrs=expected_attrs)

    @base.remotable
    def destroy(self, context):
        if not self.obj_attr_is_set('id'):
            raise exceptions.ObjectActionError(
                action='destroy', reason='no uuid')

        _destroy_vnf_deployment_flavour(context, self.id)

    def obj_load_attr(self, attrname):
        if not self._context:
            raise exceptions.OrphanedObjectError(
                method='obj_load_attr', objtype=self.obj_name())
        if 'id' not in self:
            raise exceptions.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)

        LOG.debug("Lazy-loading '%(attr)s' on %(name)s id %(id)s",
                  {'attr': attrname,
                   'name': self.obj_name(),
                   'id': self.id,
                   })

        self._obj_load_attr(attrname)

    def _obj_load_attr(self, attrname):
        """Internal method for loading attributes from vnf deployment flavour.

        """

        if attrname == 'software_images':
            self._load_sw_images()
        elif attrname in self.fields and attrname != 'id':
            self._load_generic(attrname)
        else:
            # NOTE(nirajsingh): Raise error if non existing field is
            # requested.
            raise exceptions.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)

        self.obj_reset_changes([attrname])

    def _load_generic(self, attrname):
        vnf_deployment_flavour = self.__class__.get_by_id(
            self._context, id=self.id, expected_attrs=None)
        if attrname not in vnf_deployment_flavour:
            raise exceptions.ObjectActionError(
                action='obj_load_attr',
                reason=_('loading %s requires recursion') % attrname)

        for field in self.fields:
            if field in vnf_deployment_flavour and field not in self:
                setattr(self, field, getattr(vnf_deployment_flavour, field))

    def _load_sw_images(self, db_sw_images=_NO_DATA_SENTINEL):
        if db_sw_images is _NO_DATA_SENTINEL:
            vnf_deployment_flavour = self.get_by_id(
                self._context, self.id, expected_attrs=['software_images'])
            if 'software_images' in vnf_deployment_flavour:
                self.software_images = vnf_deployment_flavour.software_images
                self.software_images.obj_reset_changes(recursive=True)
                self.obj_reset_changes(['software_images'])
            else:
                self.software_images = (
                    objects.VnfSoftwareImagesList(objects=[]))
        elif db_sw_images:
            self.software_images = base.obj_make_list(
                self._context, objects.VnfSoftwareImagesList(
                    self._context),
                objects.VnfSoftwareImage, db_sw_images)
            self.obj_reset_changes(['software_images'])

    def to_dict(self, include_fields=None):
        software_images = list()
        for software_image in self.software_images:
            sw_image_dict = software_image.to_dict(
                include_fields=include_fields)
            if sw_image_dict:
                software_images.append(sw_image_dict)

        return software_images


@base.TackerObjectRegistry.register
class VnfDeploymentFlavoursList(ovoo_base.ObjectListBase, base.TackerObject):

    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('VnfDeploymentFlavour')
    }

    def to_dict(self, include_fields=None):
        software_images = list()
        for deployment_flavour in self.objects:
            images = deployment_flavour.to_dict(include_fields=include_fields)
            if images:
                software_images.extend(images)

        return software_images



# Copyright 2016 Brocade Communications System, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
import yaml

from glanceclient import exc
from novaclient import exceptions
from oslo_config import cfg

from tacker.plugins.common import constants as evt_constants
from tacker.tests import constants
from tacker.tests.functional import base


CONF = cfg.CONF
VNF_CIRROS_CREATE_TIMEOUT = 120


class VnfTestToscaCreate(base.BaseTackerTest):
    def _test_create_vnf(self, vnfd_file, vnf_name,
                         template_source="onboarded"):

        if template_source == "onboarded":
            (vnfd_instance,
             vnf_instance,
             tosca_dict) = self.vnfd_and_vnf_create(vnfd_file, vnf_name)

        if template_source == 'inline':
            vnf_instance, tosca_dict = self.vnfd_and_vnf_create_inline(
                vnfd_file, vnf_name)

        vnfd_id = vnf_instance['vnf']['vnfd_id']
        vnf_id = vnf_instance['vnf']['id']
        self.wait_until_vnf_active(
            vnf_id,
            constants.VNF_CIRROS_CREATE_TIMEOUT,
            constants.ACTIVE_SLEEP_TIME)
        vnf_show_out = self.client.show_vnf(vnf_id)['vnf']
        self.assertIsNotNone(vnf_show_out['mgmt_ip_address'])

        prop_dict = tosca_dict['topology_template']['node_templates'][
            'CP1']['properties']

        # Verify if ip_address is static, it is same as in show_vnf
        if prop_dict.get('ip_address'):
            mgmt_ip_address_input = prop_dict.get('ip_address')
            mgmt_info = yaml.safe_load(
                vnf_show_out['mgmt_ip_address'])
            self.assertEqual(mgmt_ip_address_input, mgmt_info['VDU1'])

        # Verify anti spoofing settings
        stack_id = vnf_show_out['instance_id']
        template_dict = tosca_dict['topology_template']['node_templates']
        for field in template_dict:
            prop_dict = template_dict[field]['properties']
            if prop_dict.get('anti_spoofing_protection'):
                self.verify_antispoofing_in_stack(stack_id=stack_id,
                                                  resource_name=field)

        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE,
            evt_constants.PENDING_CREATE, cnt=2)
        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE, evt_constants.ACTIVE)
        return vnfd_id, vnf_id

    def _test_delete_vnf(self, vnf_id):
        # Delete vnf_instance with vnf_id
        try:
            self.client.delete_vnf(vnf_id)
        except Exception:
            assert False, "vnf Delete failed"

        self.wait_until_vnf_delete(vnf_id,
                                   constants.VNF_CIRROS_DELETE_TIMEOUT)
        self.verify_vnf_crud_events(vnf_id, evt_constants.RES_EVT_DELETE,
                                    evt_constants.PENDING_DELETE, cnt=2)

    def _test_create_delete_vnf_tosca(self, vnfd_file, vnf_name,
            template_source):
        vnfd_id, vnf_id = self._test_create_vnf(vnfd_file, vnf_name,
                                                template_source)
        servers = self.novaclient().servers.list()
        vdus = []
        for server in servers:
            vdus.append(server.name)
        self.assertIn('test-vdu', vdus)

        port_list = self.neutronclient().list_ports()['ports']
        vdu_ports = []
        for port in port_list:
            vdu_ports.append(port['name'])
        self.assertIn('test-cp', vdu_ports)
        self._test_delete_vnf(vnf_id)

    def test_create_delete_vnf_tosca_from_vnfd(self):
        self._test_create_delete_vnf_tosca('sample-tosca-vnfd.yaml',
                                           'test_tosca_vnf_with_cirros',
                                           'onboarded')

    def test_create_delete_vnf_from_template(self):
        self._test_create_delete_vnf_tosca('sample-tosca-vnfd.yaml',
                                           'test_tosca_vnf_with_cirros_inline',
                                           'inline')

    def test_re_create_delete_vnf(self):
        self._test_create_delete_vnf_tosca('sample-tosca-vnfd.yaml',
                                           'test_vnf',
                                           'inline')
        time.sleep(1)
        self._test_create_delete_vnf_tosca('sample-tosca-vnfd.yaml',
                                           'test_vnf',
                                           'inline')

    def test_create_delete_vnf_static_ip(self):
        vnfd_id, vnf_id = self._test_create_vnf(
            'sample-tosca-vnfd-static-ip.yaml',
            'test_tosca_vnf_with_cirros_no_monitoring')
        self._test_delete_vnf(vnf_id)


class VnfTestToscaCreateFlavorCreation(base.BaseTackerTest):
    def test_create_delete_vnf_tosca_no_monitoring(self):
        vnf_name = 'tosca_vnfd_with_auto_flavor'
        vnfd_file = 'sample-tosca-vnfd-flavor.yaml'
        vnfd_instance, vnf_instance, tosca_dict = self.vnfd_and_vnf_create(
            vnfd_file, vnf_name)
        vnf_id = vnf_instance['vnf']['id']
        self.wait_until_vnf_active(
            vnf_id,
            constants.VNF_CIRROS_CREATE_TIMEOUT,
            constants.ACTIVE_SLEEP_TIME)
        self.assertIsNotNone(self.client.show_vnf(vnf_id)['vnf'][
            'mgmt_ip_address'])

        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE,
            evt_constants.PENDING_CREATE, cnt=2)
        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE, evt_constants.ACTIVE)

        servers = self.novaclient().servers.list()
        vdu_server = None
        for server in servers:
            if 'VDU1_flavor_func' in server.name:
                vdu_server = server
                break
        self.assertIsNotNone(vdu_server)
        flavor_id = server.flavor["id"]
        nova_flavors = self.novaclient().flavors
        flavor = nova_flavors.get(flavor_id)
        self.assertIsNotNone(flavor)
        self.assertIn('VDU1_flavor_func_flavor', flavor.name)
        # Delete vnf_instance with vnf_id
        try:
            self.client.delete_vnf(vnf_id)
        except Exception:
            assert False, "vnf Delete failed"

        self.wait_until_vnf_delete(vnf_id,
                                   constants.VNF_CIRROS_DELETE_TIMEOUT)
        self.verify_vnf_crud_events(vnf_id, evt_constants.RES_EVT_DELETE,
                                    evt_constants.PENDING_DELETE, cnt=2)

        self.assertRaises(exceptions.NotFound, nova_flavors.delete,
                          [flavor_id])


class VnfTestToscaCreateImageCreation(base.BaseTackerTest):

    def test_create_delete_vnf_tosca_no_monitoring(self):
        vnf_name = 'tosca_vnfd_with_auto_image'
        vnfd_file = 'sample-tosca-vnfd-image.yaml'
        vnfd_instance, vnf_instance, tosca_dict = self.vnfd_and_vnf_create(
            vnfd_file, vnf_name)
        vnf_id = vnf_instance['vnf']['id']
        self.wait_until_vnf_active(
            vnf_id,
            constants.VNF_CIRROS_CREATE_TIMEOUT,
            constants.ACTIVE_SLEEP_TIME)
        self.assertIsNotNone(self.client.show_vnf(vnf_id)['vnf'][
            'mgmt_ip_address'])

        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE,
            evt_constants.PENDING_CREATE, cnt=2)
        self.verify_vnf_crud_events(
            vnf_id, evt_constants.RES_EVT_CREATE, evt_constants.ACTIVE)

        servers = self.novaclient().servers.list()
        vdu_server = None
        for server in servers:
            if 'VDU1_image_func' in server.name:
                vdu_server = server
                break
        self.assertIsNotNone(vdu_server)
        image_id = vdu_server.image["id"]
        glanceclient = self.glanceclient()
        image = glanceclient.images.get(image_id)
        self.assertIsNotNone(image)
        self.assertIn('VNFImage_image_func', image.name)
        # Delete vnf_instance with vnf_id
        try:
            self.client.delete_vnf(vnf_id)
        except Exception:
            assert False, "vnf Delete failed"

        self.wait_until_vnf_delete(vnf_id,
                                   constants.VNF_CIRROS_DELETE_TIMEOUT)
        self.verify_vnf_crud_events(vnf_id, evt_constants.RES_EVT_DELETE,
                                    evt_constants.PENDING_DELETE, cnt=2)

        self.assertRaises(exc.HTTPNotFound,
                          glanceclient.images.delete,
                          [image_id])



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_log import log as logging

from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.sol_refactored import objects


LOG = logging.getLogger(__name__)  # not used at the moment


def get_inst(context, inst_id):
    inst = objects.VnfInstanceV2.get_by_id(context, inst_id)
    if inst is None:
        raise sol_ex.VnfInstanceNotFound(inst_id=inst_id)
    return inst


def get_inst_all(context):
    return objects.VnfInstanceV2.get_all(context)


def inst_href(inst_id, endpoint):
    return "{}/vnflcm/v2/vnf_instances/{}".format(endpoint, inst_id)


def make_inst_links(inst, endpoint):
    links = objects.VnfInstanceV2_Links()
    self_href = inst_href(inst.id, endpoint)
    links.self = objects.Link(href=self_href)
    if inst.instantiationState == 'NOT_INSTANTIATED':
        links.instantiate = objects.Link(href=self_href + "/instantiate")
    else:  # 'INSTANTIATED'
        links.terminate = objects.Link(href=self_href + "/terminate")
        links.scale = objects.Link(href=self_href + "/scale")
        links.heal = objects.Link(href=self_href + "/heal")
        links.changeExtConn = objects.Link(href=self_href + "/change_ext_conn")
        # NOTE: add when the operation supported

    return links


# see IETF RFC 7396
def json_merge_patch(target, patch):
    if isinstance(patch, dict):
        if not isinstance(target, dict):
            target = {}
        for key, value in patch.items():
            if value is None:
                if key in target:
                    del target[key]
            else:
                target[key] = json_merge_patch(target.get(key), value)
        return target
    else:
        return patch


def select_vim_info(vim_connection_info):
    # NOTE: It is assumed that vimConnectionInfo has only one item
    # at the moment. If there are multiple items, it is uncertain
    # which item is selected.
    for vim_info in vim_connection_info.values():
        return vim_info
